"""
conftest.py is used to set up shared fixtures, plugins, hooks, other config etc.
This is the functional test conftest.py
http://docs.pytest.org/en/latest/writing_plugins.html#conftest-py-local-per-directory-plugins
"""
from __future__ import absolute_import

import os
import re
import sys
from contextlib import contextmanager

import pytest

import redfish

# For cross-compatibility between Python 2 and Python 3
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from ..conftest import RDMC_FILE
from . import utils
from .ilomock import MockServer


@pytest.fixture(scope='session')
def mock_server():
    with MockServer() as server:
        yield server


@pytest.fixture
def dl360_large(mock_server):
    """Returns a mock server config:

      - DL360 Gen10
      - 1x AEP DIMM
      - 1x DRAM DIMM
    """
    data_dir = os.path.join(os.path.dirname(__file__), 'mockdata/dl360-large')
    mock_server.load(data_dir)
    return MockServerConfig(mock_server, username='rev', password='hpe123')


@pytest.fixture
def dl360_memfast_disabled(dl360_large):
    """Returns a dl360_configured config with MemFastTraining disabled
    """
    dl360_large.patch('/redfish/v1/systems/1/bios/settings/', {'Attributes': {'MemFastTraining': 'Disabled'}})

    return dl360_large


@pytest.fixture(scope='session')
def test_server():
    """Returns a server config specified by the TEST_SERVER environment variable.

    TEST_SERVER should be a URL with optional auth, e.g. https://user:password@host:port
    """
    if 'TEST_SERVER' not in os.environ:
        pytest.skip("TEST_SERVER not set")

    url, username, password = utils.urlparse_auth(os.getenv('TEST_SERVER'))
    # If no credentials, then populate with hard-coded values. Credentials required by ilorest.
    username = "orange" if username is None else username
    password = "password" if password is None else password
    return ServerConfig(url=url, username=username, password=password)


@pytest.fixture
def ilorest(tmpdir):
    """Returns an ilorest session (in scripting mode by default).

    Each ilorest session receives its own unique temporary cache directory.
    """
    yield Ilorest(cache_dir=tmpdir.dirname)
    tmpdir.remove(ignore_errors=True)


class ServerConfig(object):

    def __init__(self, url=None, username=None, password=None):
        self.url = url
        self.username = username
        self.password = password

    @property
    def config(self):
        return self.url, self.username, self.password

    @property
    def hostname(self):
        parts = urlparse(self.url)
        hostname = parts.hostname if parts.scheme else self.url
        return hostname


class MockServerConfig(ServerConfig):

    def __init__(self, server, username=None, password=None):
        super(MockServerConfig, self).__init__(url=server.url, username=username, password=password)
        self.server = server

    def patch(self, path, data):
        """Patches a resource in the mock server."""
        self.server.patch(path, data)


class Ilorest(object):
    """Wrapper for the iLO RESTful Interface Tool."""

    def __init__(self, cache_dir=None):
        """Starts ilorest in scripting mode.

        :param cache_dir: directory to use for caching instead of the default global location
        """
        self.cache_dir = cache_dir
        self.args = [sys.executable, RDMC_FILE]

        if cache_dir:
            self.args.append('--cache-dir={}'.format(cache_dir))

        self.interactive = False
        self.proc = None
        self.prompt = "iLOrest > "

    def run(self, args, sendline=None):
        """Runs an ilorest command.

        :param args: list of command arguments
        :param sendline: string to send after running (only used in interactive mode)
        :return: results of the command as a named tuple: (stdout, stderr, retcode)
        """
        if not self.interactive:
            return utils.run_cmd(self.args + args)

        # interactive mode
        cmd_line = ' '.join(args)
        self.proc.sendline(cmd_line)
        if sendline:
            self.proc.sendline(sendline)

        stdout, stderr = self.proc.read(expect=self.prompt)

        match = re.search("iLOrest return code: ([0-9]+)\n", stdout)
        retcode = int(match.group(1)) if match else None

        return utils.CommandResult(stdout=stdout, stderr=stderr, retcode=retcode)

    def login(self, url=None, username=None, password=None):
        """Logs in to the server and establishes a secure session.

        :param url: iLO server url if connecting remotely
        :param username: iLO username. If no username specified, use "orange"
        :param password: iLO password. If no password specified, use "password"
        """
        args = ['login']
        if url:
            args.append(url)
        if username:
            args.extend(['-u', username])
        if password:
            args.extend(['-p', password])

        results = self.run(args)
        assert results.retcode == 0, "Failed to connect or log in"

    def logout(self):
        """Ends the current session and disconnects from the server."""
        self.run(['logout'])

    @contextmanager
    def session(self, url=None, username=None, password=None):
        """Establishes a session with an iLO server and logs out afterwards."""
        self.login(url=url, username=username, password=password)
        yield self
        self.logout()

    @contextmanager
    def interactive_session(self, url=None, username=None, password=None):
        """Establishes an interactive session with an iLO server, and logs out and exits afterwards."""
        with self.interactive_mode(), self.session(url=url, username=username, password=password):
            yield self

    @contextmanager
    def interactive_mode(self):
        """Enters interactive mode and exits afterwards."""
        # use --verbose to extract the return code
        self.proc = utils.InteractivePopen(self.args + ['--verbose'])
        self.proc.read(expect=self.prompt)
        self.interactive = True
        yield self
        self.proc.communicate('exit')
        self.interactive = False


class RestClient(object):
    """REST client for an iLO server."""

    def __init__(self, url=None, username=None, password=None):
        """Initializes an iLO REST client.

        :param url: iLO server url if connecting remotely
        :param username: iLO username
        :param password: iLO password
        """
        # if no scheme/protocol defined, try https
        if "://" not in url:
            url = "https://" + url
        self._client = redfish.redfish_client(base_url=url, username=username, password=password)
        self._client.login(auth=redfish.AuthMethod.BASIC)

    def get(self, path):
        """Sends a GET request."""
        resp = self._client.get(path)
        return resp.dict

    def patch(self, path, body):
        """
        Sends a PATCH request to the path with the given body
        :param path: URI path
        :param body: the body of the patch. For example dict:{ "Attributes": { "Prop1": true } }
        :return: RestResponse status, RestResponse dictionary
        """
        rest_response = self._client.patch(path=path, body=body)
        return rest_response.status, rest_response.dict



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Exit Command for rdmc """

import sys

try:
    from rdmc_helper import InvalidCommandLineErrorOPTS, ReturnCodes
except ImportError:
    from ilorest.rdmc_helper import InvalidCommandLineErrorOPTS, ReturnCodes


class ExitCommand:
    """Exit class to handle exiting from interactive mode"""

    def __init__(self):
        self.ident = {
            "name": "exit",
            "usage": None,
            "description": "Run to exit from the interactive shell\n\texample: exit",
            "summary": "Exits from the interactive shell.",
            "aliases": ["quit"],
            "auxcommands": ["LogoutCommand"],
        }
        # self.rdmc = rdmcObj
        # self.logoutobj = rdmcObj.commands_dict["LogoutCommand"](rdmcObj)

        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """If an argument is present, print help else exit

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (_, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if not args or not line:
            self.auxcommands["logout"].run("")

            # System exit
            sys.exit(ReturnCodes.SUCCESS)
        else:
            self.rdmc.ui.error("Exit command does not take any parameters.\n")
            raise InvalidCommandLineErrorOPTS("Invalid command line arguments.")

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""This is the helper class with functions that manipulate REST data"""
from __future__ import absolute_import  # check if python3 supported

import json
import multiprocessing
from multiprocessing.dummy import Pool as ThreadPool


class RestHelpers(object):
    """This is the helper class with functions that manipulate REST data"""

    def __init__(self, rdmcObject):
        self.rdmc = rdmcObject  # relies on the updated reference to the RDMC class object

    def get_resource(self, url):
        """
        Perform a GET request for the specified URL
        :param url: the URL of the resource to fetch
        :type: string
        :returns: object containing the REST responsed
        :rtype: RestResponse object
        """
        accepted_status = [200, 202]
        resp = self.rdmc.app.get_handler(url, service=True, silent=True)
        if resp and resp.status in accepted_status and resp.dict:
            return resp.dict
        return None

    def retrieve_memory_resources(self):
        """
        Retrieve the expanded Memory Collection resource
        """
        return self.get_resource("/redfish/v1/systems/1/memory?$expand=.")

    def retrieve_mem_domain_resources(self, chunks_flag=True):
        """
        Retrieve Memory Domain Resources and All Chunks
        """
        mem_domain_resources = self.get_resource("/redfish/v1/systems/1/MemoryDomains?$expand=.")
        domain_members = []
        all_chunks = []
        if mem_domain_resources:
            domain_members = mem_domain_resources.get("Members")
            if not chunks_flag:
                return domain_members
            chunk_id_list = list()
            for member in domain_members:
                chunk_id_list.append(member.get("MemoryChunks").get("@odata.id") + "?$expand=.")
            if chunk_id_list:
                chunks = self.concurrent_get(chunk_id_list)
            else:
                chunks = []
            # combining all chunks
            for chunk in chunks:
                chunk_members = chunk.get("Members")
                if chunk_members:
                    temp_chunks = [member for member in chunk_members]
                else:
                    temp_chunks = []
                all_chunks.extend(temp_chunks)
        return domain_members, all_chunks

    @staticmethod
    def filter_task_members(task_members):
        """
        Filters new memory chunk task members
        :param task_members: list of task members to be filtered
        :return: list of filtered task members
        """
        members = list()
        for member in task_members:
            # retrieving only new tasks
            if member.get("TaskState") == "New":
                target_uri = member.get("Payload").get("TargetUri")
                # checking if type Memory Chunk collection
                if "MemoryChunks" in target_uri:
                    json_body = member.get("Payload").get("JsonBody")
                    if json_body:
                        member["Payload"]["JsonBody"] = json.loads(json_body)
                    members.append(member)
        return members

    def retrieve_task_members(self):
        """
        Retrieve task members
        :returns: list of task members
        """
        # retrieving task resources
        task_resources = self.get_resource("/redfish/v1/TaskService/Tasks?$expand=.")
        # getting task members
        if task_resources:
            task_members = task_resources.get("Members")
            if task_members:
                return task_members
        return []

    def retrieve_mem_and_mem_domains(self):
        """
        Helper Function to retrieve Memory Resources and Memory Domain Resources
        concurrently
        :returns: memory resources, memory domain members and a list of memory chunks
        """
        # 'resource_list' is a list of functions that will be called to retrieve resources.
        resource_list = [
            self.retrieve_memory_resources,
            self.retrieve_mem_domain_resources,
        ]
        # 'response_list' contains the list of responses received on calling the functions
        # that are a part of 'resource_list', in the same order.
        response_list = self.concurrent_retrieve(resource_list)
        # Segregating individual responses from 'response_list':
        # The 'get()' call is responsible for retrieving the value that a function returns
        # from the 'AsyncResult' object returned by 'apply_async' in 'concurrent_retrieve'.
        memory = response_list[0].get()
        # domain_members and all chunks are returned as a tuple in 'response_list' as per the
        # 'retrieve_mem_domain_resources()' function definition. They are unpacked here.
        (domain_members, all_chunks) = response_list[1]
        # Returns the values of resources obtained above.
        return memory, domain_members, all_chunks

    def retrieve_task_members_and_mem_domains(self):
        """
        Helper Function to retrieve Task Resources and Memory Domain Resources
        concurrently
        :returns: task members, memory domain members and a list of memory chunks
        """
        # 'resource_list' is a list of functions that will be called to retrieve resources.
        resource_list = [self.retrieve_task_members, self.retrieve_mem_domain_resources]
        # 'response_list' contains the list of responses received on calling the functions
        # that are a part of 'resource_list', in the same order.
        response_list = self.concurrent_retrieve(resource_list)
        # Segregating individual responses from 'response_list':
        # The 'get()' call is responsible for retrieving the value that a function returns
        # from the 'AsyncResult' object returned by 'apply_async' in 'concurrent_retrieve'.
        task_members = response_list[0].get()
        # domain_members and all chunks are returned as a tuple in 'response_list' as per the
        # 'retrieve_mem_domain_resources()' function definition. They are unpacked here.
        (domain_members, all_chunks) = response_list[1]
        # Returns the values of resources obtained above.
        return task_members, domain_members, all_chunks

    @staticmethod
    def concurrent_retrieve(resource_list):
        """
        Concurrently retrieve Resources
        :param resource_list: List of resources to retrieve concurrently
        :type: List of functions
        :returns: List of response objects obtained as a result of GET requests
        :rtype: List of objects
        """
        response_list = list()
        # Spawn a pool of worker threads based on the CPU core count and
        # the number of functions to call.
        # 'len(resource_list) - 1' ensures that the last function call happens
        # on the main thread and not on a worker thread.
        pool = ThreadPool(min(len(resource_list) - 1, multiprocessing.cpu_count()))
        # Asynchronously call funtions from 'resource_list' on worker threads and
        # append responses to 'response_list'. These responses will be 'AsyncResult'
        # objects and the actual return value will have to be retrieved by a 'get()'
        # call to the returned response.
        for resource in resource_list[:-1]:
            response_list.append(pool.apply_async(resource))
        # Call the last function in 'resource_list' on the main thread so that
        # the main thread is doing some useful work and not sitting idle waiting
        # for all other threads to complete. This reduces the thread communication
        # overhead as well. The response returned here will not be an 'AsyncResult'
        # object and no 'get()' call is required to retrieve the actual return value.
        temp_response = resource_list[-1]()
        # Wait for all worker threads to complete execution.
        pool.close()
        pool.join()
        # Once all threads have finished execution, append the response recieved
        # on the main thread to the 'response_list'.
        response_list.append(temp_response)
        # Return list of responses obtained from the main and worker threads.
        return response_list

    def concurrent_get(self, uri_list):
        """
        Spawns a Threadpool and sends concurrent GET Requests
        :param uri_list: List of URIs on which the GET Requests
                         are supposed to be made
        :returns: List of responses from the GET Requests
        """
        # Spawn a pool of worker threads based on the CPU core count and
        # the number of GET requests to be sent
        pool = ThreadPool(min(multiprocessing.cpu_count(), len(uri_list)))
        response_list = pool.map(self.get_resource, uri_list)
        # Wait for worker threads to complete
        pool.close()
        pool.join()
        # Return list of responses obtained via GET requests from worker threads
        return response_list

    def delete_resource(self, url):
        """
        Perform a delete request for the specified URL
        :param url: the URL of the resource to delete
        :type: string
        :returns: status code
        """
        resp = self.rdmc.app.delete_handler(url, service=True, silent=True)
        if resp and resp.status in [200, 202]:
            return resp.status
        return None

    def post_resource(self, path, body):
        """
        Perform a post request for the specified path with body
        :param path: the URL path
        :type: string
        :body: the body to be sent
        :type: string
        :returns: status code
        """
        accepted_status = [200, 201, 202, 204]
        resp = self.rdmc.app.post_handler(path, body, service=True, silent=True)
        if resp and resp.status in accepted_status:
            return resp.status
        return None

    def in_post(self):
        """
        Check whether the server is in POST
        :return: True if server is in Post, False otherwise
        :rtype: Boolean
        """
        resp_body = self.get_resource("/redfish/v1/systems/1")
        if resp_body:
            if resp_body.get("Oem").get("Hpe").get("PostState") == "FinishedPost":
                return False
        return True

    def retrieve_security_state(self, path):
        """
        Get the security state
        """
        return self.get_resource(path)

    def retrieve_pmem_location(self):
        """
        Retrieve the Pmem location
        """
        return self.get_resource("/redfish/v1/Systems/1/Memory/?$expand=.#")

    def retrieve_model(self, rdmcObj):
        """
        Retrieve the Model
        """
        self.rdmc = rdmcObj
        return self.get_resource("/redfish/v1/Chassis/1")



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Create Volume Command for rdmc """

from argparse import RawDescriptionHelpFormatter

from redfish.ris import IdTokenError
from redfish.ris.rmc_helper import IloResponseError, InstanceNotFoundError

try:
    from rdmc_helper import (
        IloLicenseError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidSmartArrayConfigurationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IloLicenseError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidSmartArrayConfigurationError,
        ReturnCodes,
    )


class CreateVolumeCommand:
    """Create volume command"""

    def __init__(self):
        self.ident = {
            "name": "createvolume",
            "usage": None,
            "description": "Creates volumes on compatible HPE SSA RAID controllers\nTo view "
            "help on specific sub-commands run: createvolume <sub-command> -h\n\n"
            "NOTE: Refer http://www.hpe.com/info/scmo for additional information "
            "on creating Volumes on Gen11 servers.\n\t"
            "Also, when you select multiple physicaldrives you can select by both\n\t"
            "physical drive name and by the location at the same time.\n\t"
            "You can also select controllers by slot number as well as index.\n\t"
            "For iLO6, storage id need to be specified using --storageid=DE00E000 "
            "along with --controller=1",
            "summary": "Creates a new volume on the selected controller.",
            "aliases": ["createlogicaldrive"],
            "auxcommands": ["SelectCommand", "StorageControllerCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main disk inventory worker function

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.createvolumevalidation(options)
        ilo_ver = self.rdmc.app.getiloversion()
        if ilo_ver < 6.110:
            try:
                self.auxcommands["select"].selectfunction("StorageController.")
                content = self.rdmc.app.getprops()
                if options.command == "quickdrive":
                    raise InvalidCommandLineError(
                        "This controller is not compatible with quickdrive option. "
                        "Please use customdrive or volume option\n"
                    )
                ilo_ver = 6.110
                options.command = "volume"
            except Exception:
                pass

        if ilo_ver >= 6.110:
            if options.command == "customdrive" or options.command == "quickdrive":
                raise InvalidCommandLineError("customdrive or quickdrive subcommand is not supported on iLO6(Gen11).\n")
            if not options.storageid:
                raise InvalidCommandLineError("--storageid option is mandatory for iLO6 or latest iLO5 onwards.\n")
        else:
            if options.command == "volume":
                raise InvalidCommandLineError("volume subcommand is not supported on iLO5(Gen10).\n")
        if options.controller:
            if ilo_ver >= 6.110:
                if options.storageid:
                    controllers = self.auxcommands["storagecontroller"].storagecontroller(options, single_use=True)
                else:
                    raise InvalidCommandLineError("--storageid option is mandatory for iLO6 onwards.\n")
            else:
                try:
                    controllers = self.auxcommands["storagecontroller"].controllers(options, single_use=True)
                    if len(controllers) == 0:
                        if not options.storageid:
                            raise InvalidCommandLineError(
                                "--storageid option is mandatory. Please input storageid as well so that "
                                "controllers/volumes can be identified.\n"
                            )
                    controllers = self.auxcommands["storagecontroller"].storagecontroller(options, single_use=True)
                    options.command = "volume"
                    ilo_ver = 6.110
                except InstanceNotFoundError:
                    if not options.storageid:
                        raise InvalidCommandLineError(
                            "--storageid option is mandatory. Please input storageid as well so that "
                            "controllers/volumes can be identified.\n"
                        )
                    controllers = self.auxcommands["storagecontroller"].storagecontroller(options, single_use=True)
                    options.command = "volume"
                    ilo_ver = 6.110
            try:
                # logical_drives = self.rdmc.app.get_handler(
                #     "/redfish/v1/Systems/1/Storage/" + options.storageid + "/Volumes/",
                #     silent=True,
                #     service=True,
                # ).dict["Members"]
                # for volume_list in logical_drives:
                #     volume = self.rdmc.app.get_handler(volume_list["@odata.id"], silent=True, service=True).dict
                #     if (
                #         ("Status" in volume)
                #         and (volume["Status"]["State"] == "Enabled")
                #         and ("RAIDType" in volume)
                #         and (volume["RAIDType"] == "None")
                #     ):
                # self.rdmc.app.delete_handler(volume["@odata.id"], {}, silent=True, service=True)

                controller = controllers[next(iter(controllers))]
                (create_flag, newdrive) = self.createvolume(options, controller)
                if create_flag:
                    if ilo_ver >= 6.110:
                        temp_odata = controller["@odata.id"]
                        volume_path = temp_odata.split("Controllers")[0] + "Volumes"
                        self.rdmc.ui.printer("CreateVolume path and payload: %s, %s\n" % (volume_path, newdrive))
                        result = self.rdmc.app.post_handler(volume_path, newdrive)

                        self.rdmc.ui.printer("Volume created successfully  \n")
                        if options.sparedrives:
                            controller["physical_drives"] = self.auxcommands[
                                "storagecontroller"
                            ].storagephysical_drives(options, controller, options.storageid, single_use=True)
                            array = options.disks.split(",")
                            set_spare = False
                            newdrive1 = {"Links": {"DedicatedSpareDrives": [{}]}}
                            volume_path1 = volume_path + result.session_location.split("Volumes")[1]

                            sparedrives = options.sparedrives.split(",")
                            s_array = []
                            if len(controller["physical_drives"]) > 0:
                                for p_id in controller["physical_drives"]:
                                    p_loc = self.convertloc(
                                        controller["physical_drives"][str(p_id)]["PhysicalLocation"]["PartLocation"][
                                            "ServiceLabel"
                                        ]
                                    )
                                    l_loc = p_loc.split(":")[1:]
                                    p_loc = ":".join(l_loc)

                                    for sdrive in sparedrives:
                                        if sdrive == p_loc and sdrive not in array:
                                            s_array.append(
                                                {
                                                    "@odata.id": controller["physical_drives"][str(p_id)]["@odata.id"],
                                                }
                                            )
                                            set_spare = True
                            newdrive1["Links"]["DedicatedSpareDrives"] = s_array
                            if not set_spare:
                                raise InvalidCommandLineError(
                                    "Invalid spare drive given, check whether given spare drive is different from "
                                    "disks given or check whether spare drive belongs to same controller"
                                )

                            self.rdmc.app.patch_handler(volume_path1, newdrive1)
                            self.rdmc.ui.printer("Successfully added the sparedrives %s\n" % (newdrive1))

                        return ReturnCodes.SUCCESS

                    else:
                        temp_odata = controller["@odata.id"]
                        payload_dict = dict()
                        payload_dict["DataGuard"] = "Disabled"
                        if "settings" not in temp_odata:
                            temp_odata = temp_odata + "settings/"
                        settings_controller = self.rdmc.app.get_handler(temp_odata, service=False, silent=True)
                        # Fix for multiple logical creation at single reboot
                        if self.rdmc.app.typepath.defs.isgen9:
                            payload_dict["logical_drives"] = dict()
                            payload_dict["logical_drives"]["new"] = newdrive
                        else:
                            payload_dict["LogicalDrives"] = settings_controller.dict["LogicalDrives"]
                            payload_dict["LogicalDrives"].append(newdrive)
                        self.rdmc.ui.printer("CreateVolume path and payload: %s, %s\n" % (temp_odata, payload_dict))
                        self.rdmc.app.put_handler(
                            temp_odata,
                            payload_dict,
                            headers={"If-Match": self.getetag(temp_odata)},
                        )
                        self.rdmc.app.download_path([temp_odata], path_refresh=True, crawl=False)
                        self.rdmc.ui.printer(
                            "One or more properties were changed and will not take effect " "until system is reset \n"
                        )
            except IloLicenseError:
                self.rdmc.ui.error("License Error Occured while creating volume\n")
                return ReturnCodes.ILO_LICENSE_ERROR
            except IdTokenError:
                self.rdmc.ui.error("Insufficient Privilege to create volume\n")
                return ReturnCodes.RIS_MISSING_ID_TOKEN
            except IloResponseError:
                self.rdmc.ui.error("iLO threw iLOResponseError\n")
                return ReturnCodes.RIS_ILO_RESPONSE_ERROR
            except StopIteration:
                self.rdmc.ui.error("Drive or Controller not exist, Please check drive or controller\n")
            except Exception as excp:
                self.rdmc.ui.error(excp)

        else:
            self.rdmc.ui.error("Provide the controller \n")

        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def convertloc(self, servicelabel):
        loc = servicelabel.split(":")
        temp_str = str(
            loc[0].split("=")[1] + ":" + loc[1].split("=")[1] + ":" + loc[2].split("=")[1] + ":" + loc[3].split("=")[1]
        )
        return temp_str

    def get_allowed_list(self, storage_id, attr):
        url = "/redfish/v1/Systems/1/Storage/" + storage_id + "/Volumes/Capabilities"
        attr_allowed_str = attr + "@Redfish.AllowableValues"

        results = self.rdmc.app.get_handler(url, service=True, silent=True).dict
        return results[attr_allowed_str]

    def createvolume(self, options, controller):
        """Create volume"""
        global p_loc
        ilo_ver = self.rdmc.app.getiloversion()
        if options.command == "volume":
            ilo_ver = 6.110
        if ilo_ver >= 6.110:
            try:
                raidlvllist = self.get_allowed_list(options.storageid, "RAIDType")
            except:
                raidlvllist = [
                    "Raid0",
                    "Raid1",
                    "Raid1ADM",
                    "Raid10",
                    "Raid10ADM",
                    "Raid5",
                    "Raid50",
                    "Raid6",
                    "Raid60",
                ]

        else:
            raidlvllist = [
                "Raid0",
                "Raid1",
                "Raid1ADM",
                "Raid10",
                "Raid10ADM",
                "Raid5",
                "Raid50",
                "Raid6",
                "Raid60",
            ]
        interfacetypelist = ["SAS", "SATA", "NVMe"]
        mediatypelist = ["SSD", "HDD"]
        sparetypelist = ["Dedicated", "Roaming"]
        acceltypelist = ["ControllerCache", "IOBypass", "None"]
        locationtypelist = ["Internal", "External"]
        legacylist = ["Primary", "Secondary", "All", "None"]
        paritylist = ["Default", "Rapid"]
        iOPerfModeEnabledlist = ["true", "false"]
        if ilo_ver >= 6.110:
            try:
                readCachePolicylist = self.get_allowed_list(options.storageid, "ReadCachePolicy")
                # writeCachePolicylist = self.get_allowed_list(options.storageid, "WriteCachePolicy")
                writeCachePolicylist = [
                    "Off",
                    "WriteThrough",
                    "ProtectedWriteBack",
                    "UnprotectedWriteBack",
                ]
            except:
                readCachePolicylist = ["Off", "ReadAhead"]
                writeCachePolicylist = [
                    "Off",
                    "WriteThrough",
                    "ProtectedWriteBack",
                    "UnprotectedWriteBack",
                ]
            # InitMethodlist = self.get_allowed_list(options.storageid, "InitializeMethod")
        WriteHoleProtectionPolicyList = ["Yes", "No"]
        sparedrives = []
        changes = False

        if ilo_ver >= 6.110:
            controller["physical_drives"] = self.auxcommands["storagecontroller"].storagephysical_drives(
                options, controller, options.storageid, single_use=True
            )
        else:
            try:
                controller["physical_drives"] = self.auxcommands["storagecontroller"].physical_drives(
                    options, controller, single_use=True
                )
            except:
                controller["physical_drives"] = self.auxcommands["storagecontroller"].storagephysical_drives(
                    options, controller, options.storageid, single_use=True
                )
        if ilo_ver >= 6.110:
            controller["logical_drives"] = self.auxcommands["storagecontroller"].storagelogical_drives(
                options, controller, options.storageid, single_use=True
            )
        else:
            try:
                controller["logical_drives"] = self.auxcommands["storagecontroller"].logical_drives(
                    options, controller, single_use=True
                )
            except:
                controller["logical_drives"] = self.auxcommands["storagecontroller"].storagelogical_drives(
                    options, controller, options.storageid, single_use=True
                )
        if controller.get("Links"):
            if ilo_ver >= 6.110:
                newdrive = {"Links": {"Drives": {}}}
            else:
                newdrive = {"Links": {"DataDrives": {}}}
        else:
            if ilo_ver >= 6.110:
                newdrive = {"Links": {"Drives": {}}}
            else:
                newdrive = {"links": {"DataDrives": {}}}

        changes = False
        itemadded = False

        for item in raidlvllist:
            if options.raid.lower() == item.lower():
                if options.command == "customdrive" or options.command == "volume":
                    drivecount = len(options.disks.replace(", ", ",").split(","))
                else:
                    try:
                        drivecount = int(options.disks)
                    except ValueError:
                        raise InvalidCommandLineError("Number of drives is not an integer.")
                if self.raidvalidation(item.lower(), drivecount, options):
                    itemadded = True
                    if ilo_ver >= 6.110:
                        newdrive["RAIDType"] = options.raid.upper()
                    else:
                        newdrive["Raid"] = item
                break

        if not itemadded:
            raise InvalidCommandLineError("Invalid raid type or configuration.")
        else:
            itemadded = False

        if options.command == "customdrive":
            if options.sparedrives:
                sparedrives = options.sparedrives.replace(", ", ",").split(",")
                newdrive["SpareDrives"] = []
                newdrive["SpareRebuildMode"] = "Dedicated"

            drives = options.disks.replace(", ", ",").split(",")
            newdrive["DataDrives"] = []

            if len(controller["physical_drives"]) > 0:
                for id in controller["physical_drives"]:
                    for drive in drives:
                        drv_id = controller["physical_drives"][str(id)]
                        if "Location" in drv_id:
                            drv_loc = drv_id["Location"]
                        else:
                            location = drv_id["PhysicalLocation"]["PartLocation"]["ServiceLabel"]
                            loc = location.split(":")
                            del loc[0]
                            if len(loc) == 3:
                                temp_str = str(
                                    loc[0].split("=")[1] + ":" + loc[1].split("=")[1] + ":" + loc[2].split("=")[1]
                                )
                                drv_loc = temp_str

                        if drive == drv_loc:
                            newdrive["DataDrives"].append(drive)

                    for sdrive in sparedrives:
                        drv_id = controller["SpareDrives"][str(id)]
                        if "Location" in drv_id:
                            drv_loc = drv_id["Location"]
                        else:
                            location = drv_id["PhysicalLocation"]["PartLocation"]["ServiceLabel"]
                            loc = location.split(":")
                            del loc[0]
                            if len(loc) == 3:
                                temp_str = str(
                                    loc[0].split("=")[1] + ":" + loc[1].split("=")[1] + ":" + loc[2].split("=")[1]
                                )
                                drv_loc = temp_str

                        if drive == drv_loc:
                            newdrive["DataDrives"].append(drive)
                        if sdrive == drv_id:
                            newdrive["SpareDrives"].append(sdrive)
            else:
                raise InvalidCommandLineError("No Physical Drives in this controller")

            if drivecount > len(newdrive["DataDrives"]):
                raise InvalidCommandLineError(
                    "Not all of the selected drives could " "be found in the specified locations."
                )

            if options.sparetype:
                itemadded = False
                for item in sparetypelist:
                    if options.sparetype.lower() == item.lower():
                        newdrive["SpareRebuildMode"] = item
                        itemadded = True
                        break

                if not itemadded:
                    raise InvalidCommandLineError("Invalid spare drive type.")

            if options.drivename:
                newdrive["LogicalDriveName"] = options.drivename

            if options.capacitygib:
                try:
                    capacitygib = int(options.capacitygib)
                except ValueError:
                    raise InvalidCommandLineError("Capacity is not an integer.")
                newdrive["CapacityGiB"] = capacitygib

            if options.acceleratortype:
                itemadded = False
                for item in acceltypelist:
                    if options.acceleratortype.lower() == item.lower():
                        newdrive["Accelerator"] = item
                        itemadded = True
                        break

                if not itemadded:
                    raise InvalidCommandLineError("Invalid accelerator type.")

            if options.legacyboot:
                itemadded = False
                for item in legacylist:
                    if options.legacyboot.lower() in item.lower():
                        newdrive["LegacyBootPriority"] = item
                        itemadded = True
                        break

                if not itemadded:
                    raise InvalidCommandLineError("Invalid legacy boot priority.")

            if options.capacityblocks:
                try:
                    capacityblocks = int(options.capacityblocks)
                except ValueError:
                    raise InvalidCommandLineError("Capacity is not an integer.")

                newdrive["CapacityBlocks"] = capacityblocks

            if options.paritygroup:
                try:
                    paritygroup = int(options.paritygroup)
                except ValueError:
                    raise InvalidCommandLineError("Parity group is not an integer.")

                newdrive["ParityGroupCount"] = paritygroup

            if options.paritytype:
                itemadded = False
                for item in paritylist:
                    if options.paritytype.lower() == item.lower():
                        newdrive["ParityInitializationType"] = item
                        itemadded = True
                        break

                if not itemadded:
                    raise InvalidCommandLineError("Invalid parity type")

            if options.blocksize:
                try:
                    blocksize = int(options.blocksize)
                except ValueError:
                    raise InvalidCommandLineError("Block size is not an integer.")

                newdrive["BlockSizeBytes"] = blocksize

            if options.stripsize:
                try:
                    stripsize = int(options.stripsize)
                except ValueError:
                    raise InvalidCommandLineError("Strip size is not an integer.")

                newdrive["StripSizeBytes"] = stripsize

            if options.stripesize:
                try:
                    stripesize = int(options.stripesize)
                except ValueError:
                    raise InvalidCommandLineError("Stripe size is not an integer.")

                newdrive["StripeSizeBytes"] = stripesize
        elif options.command == "quickdrive":
            try:
                numdrives = int(options.disks)
            except ValueError:
                raise InvalidCommandLineError("Number of drives is not an integer.")

            newdrive["DataDrives"] = {
                "DataDriveCount": numdrives,
                "DataDriveMinimumSizeGiB": 0,
            }
            for item in mediatypelist:
                if options.drivetype.lower() == item.lower():
                    newdrive["DataDrives"]["DataDriveMediaType"] = item
                    itemadded = True
                    break
            if not itemadded:
                raise InvalidCommandLineError("Invalid media type.")
            else:
                itemadded = False
            for item in interfacetypelist:
                if options.interfacetype.lower() == item.lower():
                    newdrive["DataDrives"]["DataDriveInterfaceType"] = item
                    itemadded = True
                    break
            if not itemadded:
                raise InvalidCommandLineError("Invalid interface type.")

            if options.locationtype:
                for item in locationtypelist:
                    if options.locationtype.lower() == item.lower():
                        newdrive["DataDrives"]["DataDriveLocation"] = item
                        break
            if options.minimumsize:
                try:
                    minimumsize = int(options.minimumsize)
                except ValueError:
                    raise InvalidCommandLineError("Minimum size is not an integer.")
                newdrive["DataDrives"]["DataDriveMinimumSizeGiB"] = minimumsize
                newdrive["CapacityGiB"] = minimumsize
        elif options.command == "volume":
            idval = []
            if len(controller["physical_drives"]) > 0:
                array = options.disks.split(",")
                DA = ["DA000000", "DA000001"]
                for p_id in controller["physical_drives"]:
                    if p_id not in DA:
                        p_loc = self.convertloc(
                            controller["physical_drives"][str(p_id)]["PhysicalLocation"]["PartLocation"]["ServiceLabel"]
                        )

                        for ar in array:
                            if ar in p_loc:
                                idval.append(controller["physical_drives"][str(p_id)]["@odata.id"])
            newdrive["Links"]["Drives"] = []
            if idval is not None:
                for id in idval:
                    newdrive["Links"]["Drives"].append(
                        {
                            "@odata.id": id,
                        }
                    )
            else:
                raise InvalidCommandLineError(
                    "Disk location given is invalid , Kindly recheck and provide valid location"
                )

            if "capacitygib" in options and options.capacitygib:
                options.capacitybytes = int(options.capacitygib) * 1024 * 1024 * 1024
            if "capacitybytes" in options and options.capacitybytes is not None:
                if options.capacitybytes:
                    try:
                        if isinstance(options.capacitybytes, list):
                            capacitybytes = int(options.capacitybytes[0])
                        else:
                            capacitybytes = int(options.capacitybytes)
                    except ValueError:
                        raise InvalidCommandLineError("Capacity is not an integer.")

                    newdrive["CapacityBytes"] = capacitybytes

            if "iOPerfModeEnabled" in options and options.iOPerfModeEnabled:
                for item in iOPerfModeEnabledlist:
                    if (
                        options.iOPerfModeEnabled[0].lower() == item.lower()
                        and options.iOPerfModeEnabled[0].lower() == "false"
                    ):
                        newdrive["IOPerfModeEnabled"] = eval(options.iOPerfModeEnabled[0])
                        itemadded = True
                        break
                    elif (
                        options.iOPerfModeEnabled[0].lower() == item.lower()
                        and options.iOPerfModeEnabled[0].lower() == "true"
                    ):
                        if "SSD" in controller["SupportedDeviceProtocols"]:
                            newdrive["IOPerfModeEnabled"] = eval(options.iOPerfModeEnabled[0])
                            itemadded = True
                            break
                        else:
                            raise InvalidCommandLineError(
                                " IOPerfModeEnabled can be true only when supported protocol is SSD"
                            )
                if not itemadded:
                    raise InvalidCommandLineError("Invalid IOPerfModeEnabled, Value should be either False or True")
                else:
                    itemadded = False

            if "ReadCachePolicy" in options and options.ReadCachePolicy is not None:
                for item in readCachePolicylist:
                    if options.ReadCachePolicy.lower() == item.lower():
                        newdrive["ReadCachePolicy"] = item
                        itemadded = True
                        break
                if not itemadded:
                    raise InvalidCommandLineError("Invalid ReadCachePolicy, Value should be 'Off' or 'ReadAhead'")
                else:
                    itemadded = False

            if "WriteCachePolicy" in options and options.WriteCachePolicy is not None:
                for item in writeCachePolicylist:
                    if options.WriteCachePolicy.lower() == item.lower():
                        newdrive["WriteCachePolicy"] = item
                        itemadded = True
                        break
                if not itemadded:
                    raise InvalidCommandLineError(
                        "Invalid WriteCachePolicy, Values should be 'Off', 'WriteThrough','ProtectedWriteBack',"
                        "'UnprotectedWriteBack'"
                    )
                else:
                    itemadded = False

            if "VROC" in controller["Model"]:
                if options.WriteHoleProtectionPolicy is not None:
                    for item in WriteHoleProtectionPolicyList:
                        if (
                            options.WriteHoleProtectionPolicy[0].lower() == item.lower()
                            and options.WriteHoleProtectionPolicy[0].lower() == "yes"
                        ):
                            newdrive["WriteHoleProtectionPolicy"] = "Journaling"
                            newdrive["Links"]["JournalingMedia"] = idval
                            itemadded = True
                            break
                    if not itemadded:
                        raise InvalidCommandLineError(
                            "Invalid WriteHoleProtectionPolicy, Values can be either Yes or No"
                        )
                    else:
                        itemadded = False
            if "drivename" in options and options.drivename:
                options.DisplayName = options.drivename
            if "DisplayName" in options and options.DisplayName is not None:
                newdrive["DisplayName"] = options.DisplayName

        if newdrive:
            if options.command == "quickdrive":
                newdrive_count = newdrive["DataDrives"]["DataDriveCount"]
            elif options.command == "customdrive":
                newdrive_count = len(newdrive["DataDrives"])
            elif options.command == "volume":
                newdrive_count = len(newdrive["Links"]["Drives"])

            if len(controller["physical_drives"]) >= newdrive_count:
                drives_avail = len(controller["physical_drives"])
                accepted_drives = 0
                for cnt, drive in enumerate(controller["physical_drives"]):
                    drivechecks = (False, False, False, False)
                    if drives_avail < newdrive_count:
                        raise InvalidSmartArrayConfigurationError(
                            "Unable to continue, requested "
                            "configuration not possible with current physical drive inventory.\n"
                        )
                    else:
                        drivechecks = (True, False, False, False)

                    if options.command == "quickdrive":
                        if (
                            controller["physical_drives"][drive]["InterfaceType"]
                            == newdrive["DataDrives"]["DataDriveInterfaceType"]
                        ):
                            drivechecks = (True, True, False, False)
                        else:
                            drives_avail -= 1
                            continue
                        if (
                            controller["physical_drives"][drive]["MediaType"]
                            == newdrive["DataDrives"]["DataDriveMediaType"]
                        ):
                            drivechecks = (True, True, True, False)
                        else:
                            drives_avail -= 1
                            continue
                    else:
                        drivechecks = (True, True, True, False)
                    in_use = False
                    if controller["logical_drives"] is not None:
                        for existing_logical_drives in controller["logical_drives"]:
                            _logical_drive = controller["logical_drives"][existing_logical_drives]
                            if _logical_drive.get("LogicalDrives"):
                                for _data_drive in _logical_drive["LogicalDrives"]["DataDrives"]:
                                    if drive == _logical_drive["LogicalDrives"]["DataDrives"][_data_drive]:
                                        in_use = True
                            elif _logical_drive.get("Links"):
                                if not ilo_ver >= 6.110:
                                    for _data_drive in _logical_drive["Links"]["DataDrives"]:
                                        if drive == _logical_drive["Links"]["DataDrives"][_data_drive]:
                                            in_use = True
                                else:
                                    for _data_drive in _logical_drive["Links"]["Drives"]:
                                        if drive == _data_drive:
                                            in_use = True
                            elif _logical_drive.get("links"):
                                for _data_drive in _logical_drive["links"]["DataDrives"]:
                                    if drive == _logical_drive["links"]["DataDrives"][_data_drive]:
                                        in_use = True
                    if in_use:
                        drives_avail -= 1
                        continue
                    else:
                        drivechecks = (True, True, True, True)
                    if drivechecks[0] and drivechecks[1] and drivechecks[2]:
                        if controller.get("Links"):
                            newdrive["Links"]["Drives"][int(drive)] = {
                                "@odata.id": controller["physical_drives"][drive]["@odata.id"]
                            }
                        else:
                            if not ilo_ver >= 6.110:
                                newdrive["links"]["DataDrives"][int(drive)] = controller["physical_drives"][drive]
                        accepted_drives += 1
                        changes = True
                        if accepted_drives == newdrive_count:
                            break
                    else:
                        drives_avail -= 1

            if changes:
                if self.rdmc.app.typepath.defs.isgen9:
                    controller["logical_drives"]["new"] = newdrive
                else:
                    if not ilo_ver >= 6.110:
                        try:
                            newdrive.pop("Links")
                        except KeyError:
                            newdrive.pop("links")
                    if not ilo_ver >= 6.110:
                        controller["LogicalDrives"].append(newdrive)
                    del controller["logical_drives"]
                    del controller["physical_drives"]

        return (changes, newdrive)

    def getetag(self, path):
        """get etag from path"""
        etag = None
        instance = self.rdmc.app.monolith.path(path)
        if instance:
            etag = (
                instance.resp.getheader("etag")
                if "etag" in instance.resp.getheaders()
                else instance.resp.getheader("ETag")
            )
        return etag

    def raidvalidation(self, raidtype, numdrives, options):
        """vaidation function for raid levels
        :param raidtype: raid type
        :type options: string.
        :param numdrives: number of drives
        :type numdrives: int.
        :param options: command line options
        :type options: list.
        """

        valid = True

        if raidtype == "raid5":
            if numdrives < 3:  # or options.stripsize:
                valid = False
        elif raidtype == "raid6":
            if numdrives < 4:  # or options.stripsize:
                valid = False
        elif raidtype == "raid50":
            if numdrives < 6:
                valid = False
        elif raidtype == "raid60":
            if numdrives < 8:
                valid = False

        return valid

    def createvolumevalidation(self, options):
        """Create volume validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    @staticmethod
    def options_argument_group(parser):
        """Define optional arguments group

        :param parser: The parser to add the login option group to
        :type parser: ArgumentParser/OptionParser
        """
        group = parser.add_argument_group(
            "GLOBAL OPTIONS",
            "Options are available for all" "arguments within the scope of this command.",
        )

        group.add_argument(
            "--controller",
            dest="controller",
            help="Use this flag to select the corresponding controller "
            "using either the slot number or index.\nexample: --controller=Slot 0 OR "
            "--controller=1",
            default=None,
            required=True,
        )

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        # self.options_argument_group(customparser)

        subcommand_parser = customparser.add_subparsers(dest="command")
        subcommand_parser.required = True
        qd_help = (
            "Create a volume with a minimal number of arguments (utilizes default "
            "values on the controller). This option is only for iLO5 or Gen10"
        )
        # quickdrive sub-parser
        qd_parser = subcommand_parser.add_parser(
            "quickdrive",
            help=qd_help,
            description=qd_help + "\n\texample: createvolume quickdrive "
            "<raid-level> <num-drives> <media-type> <interface-type> "
            "--locationtype=Internal  --minimumsize=0 --controller=1",
            formatter_class=RawDescriptionHelpFormatter,
        )
        qd_parser.add_argument(
            "raid",
            help="Specify the RAID level for the volume to be created.",
            metavar="Raid_Level",
        )
        qd_parser.add_argument(
            "disks",
            help="For quick drive creation, specify number of disks.",
            metavar="Drives",
        )
        qd_parser.add_argument(
            "drivetype",
            help="Specify the drive media type of the physical disk(s) (i.e. HDD or SSD)",
            metavar="Drive_Media_Type",
        )
        qd_parser.add_argument(
            "interfacetype",
            help="Specify the interface type of the physical disk(s) (i.e. SATA or SAS or NVMe)",
            metavar="Drive_Interface_Type",
        )
        qd_parser.add_argument(
            "--locationtype",
            dest="locationtype",
            help="Optionally specify the location of the physical disks(s) (i.e. Internal or External)",
            default=None,
        )
        qd_parser.add_argument(
            "--minimumsize",
            dest="minimumsize",
            help="""Optionally include to set the minimum size of the drive """
            """in GiB. (usable in quick creation only, use -1 for max size)""",
            default=None,
        )
        qd_parser.add_argument(
            "--controller",
            dest="controller",
            help="Use this flag to select the corresponding controller "
            "using either the slot number or index.\nexample: --controller=Slot 0 OR "
            "--controller=1",
            default=None,
            required=True,
        )
        qd_parser.add_argument(
            "--storageid",
            dest="storageid",
            help="Use this flag to select the corresponding storageid "
            "using either the slot number or index.\nexample: --storageid=DE123234",
            default=None,
            required=False,
        )
        self.cmdbase.add_login_arguments_group(qd_parser)
        # self.options_argument_group(qd_parser)

        cd_help = (
            "Create a customised volume using all available properties (as optional "
            "arguments) for creation. This option is only for iLO5 or Gen10"
        )
        # customdrive sub-parser
        cd_parser = subcommand_parser.add_parser(
            "customdrive",
            help=cd_help,
            description=cd_help + "\n\texample: createvolume customdrive "
            "<raid-level> <physicaldrivelocations> --controller=1 "
            "--name=drivename --spare-drives=1I:1:1,1I:1:3 "
            "--spare-type=Dedicated --capacitygib=10 "
            "--accelerator-type=None\n\n\tOPTIONS:\n\traid-level:\t\t"
            "Raid0, Raid1, Raid1ADM, Raid10, Raid10ADM, Raid5, Raid50, "
            "Raid6, Raid60\n\tphysicaldrivelocation(s):\tLocation, Drive-name\n\t"
            "media-type:\t\tSSD,HDD\n\tinterface-type:"
            "\t\tSAS, SATA, NVMe\n\tdrive-location:\t\tInternal, External\n\t"
            "--spare-type:\t\tDedicated, Roaming\n\t--accelerator-type:\t"
            "ControllerCache, IOBypass, None\n\t--paritytype:\t\tDefault, Rapid"
            "\n\t--capacitygib:\t\t-1 (for Max Size)\n\t--capacityblocks:\t"
            "-1 (for Max Size)\n\n\t",
            formatter_class=RawDescriptionHelpFormatter,
        )
        cd_parser.add_argument(
            "raid",
            help="Specify the RAID level for the volume to be created.",
            metavar="Raid_Level",
        )
        cd_parser.add_argument(
            "disks",
            help="For custom drive, specify a comma separated physical disk locations.",
            metavar="Drive_Indices",
        )
        cd_parser.add_argument(
            "-n",
            "--name",
            dest="drivename",
            help="""Optionally include to set the drive name (usable in """ """custom creation only).""",
            default=None,
        )
        cd_parser.add_argument(
            "--spare-drives",
            dest="sparedrives",
            help="""Optionally include to set the spare drives by the """
            """physical drive's location. (usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--capacitygib",
            dest="capacitygib",
            help="""Optionally include to set the capacity of the drive in """
            """GiB. (usable in custom creation only, use -1 for max """
            """size)""",
            default=None,
        )
        cd_parser.add_argument(
            "--accelerator-type",
            dest="acceleratortype",
            help="""Optionally include to choose the accelerator type.""",
            default=None,
        )
        cd_parser.add_argument(
            "--spare-type",
            dest="sparetype",
            help="""Optionally include to choose the spare drive type. """ """(usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--minimumsize",
            dest="minimumsize",
            help="""Optionally include to set the minimum size of the drive """
            """in GiB. (usable in quick creation only, use -1 for max size)""",
            default=None,
        )
        cd_parser.add_argument(
            "--legacy-boot",
            dest="legacyboot",
            help="""Optionally include to choose the legacy boot priority. """ """(usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--storageid",
            dest="storageid",
            help="Use this flag to select the corresponding storageid "
            "using either the slot number or index.\nexample: --storageid=DE123234",
            default=None,
            required=False,
        )
        cd_parser.add_argument(
            "--capacityblocks",
            dest="capacityblocks",
            help="""Optionally include to choose the capacity in blocks. """
            """(use -1 for max size, usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--paritygroupcount",
            dest="paritygroup",
            help="""Optionally include to include the number of parity """
            """groups to use. (only valid for certain RAID levels)""",
            default=None,
        )
        cd_parser.add_argument(
            "--paritytype",
            dest="paritytype",
            help="""Optionally include to choose the parity initialization"""
            """ type. (usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--block-size-bytes",
            dest="blocksize",
            help="""Optionally include to choose the block size of the disk"""
            """ drive. (usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--strip-size-bytes",
            dest="stripsize",
            help="""Optionally include to choose the strip size in bytes. """ """(usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--stripe-size-bytes",
            dest="stripesize",
            help="""Optionally include to choose the stripe size in bytes. """ """(usable in custom creation only)""",
            default=None,
        )
        cd_parser.add_argument(
            "--controller",
            dest="controller",
            help="Use this flag to select the corresponding controller "
            "using either the slot number or index.\nexample: --controller=Slot 0 OR "
            "--controller=1",
            default=None,
            required=True,
        )
        self.cmdbase.add_login_arguments_group(cd_parser)
        # self.options_argument_group(cd_parser)
        v_help = (
            "Create a volume using all available properties (as optional "
            "arguments) for creation on gen 11 or higher "
        )
        # volume sub-parser
        v_parser = subcommand_parser.add_parser(
            "volume",
            help=v_help,
            description=v_help + "\n\texample: createvolume volume "
            "<raid-level> <physicaldrivelocations> <displayname> "
            "<iOPerfModeEnabled> <readCachePolicy> "
            "<writeCachePolicy> <WriteHoleProtectionPolicy> --storageid=DE009000 --controller=0 "
            "<spare-drives> <capacitygib> "
            "\n\n\t",
            formatter_class=RawDescriptionHelpFormatter,
        )
        v_parser.add_argument(
            "raid",
            help="Specify the RAID level for the volume to be created.",
            metavar="Raid_Level",
        )
        v_parser.add_argument(
            "disks",
            help="For custom drive, specify a comma separated physical disk locations.",
            metavar="Drive_Indices",
        )
        v_parser.add_argument(
            "--storageid",
            dest="storageid",
            help="Use this flag to select the corresponding storageid "
            "using either the slot number or index.\nexample: --storageid=DE123234",
            default=None,
            required=True,
        )
        v_parser.add_argument(
            "--DisplayName",
            dest="DisplayName",
            help="""Optionally include to set the drive name """,
            default=None,
        )
        v_parser.add_argument(
            "--iOPerfModeEnabled",
            dest="iOPerfModeEnabled",
            help="""Optionally include to choose the IOPerfModeEnabled . Allowed values are 'True', 'False'""",
            default=None,
        )
        v_parser.add_argument(
            "--ReadCachePolicy",
            dest="ReadCachePolicy",
            help="""Optionally include to choose the ReadCachePolicy. """ """Allowed values are 'Off', 'ReadAhead'""",
            default=None,
        )
        v_parser.add_argument(
            "--WriteCachePolicy",
            dest="WriteCachePolicy",
            help="Optionally include to set the WriteCachePolicy "
            "Allowed values are 'Off', 'WriteThrough','ProtectedWriteBack','UnprotectedWriteBack'",
            default=None,
        )
        v_parser.add_argument(
            "--WriteHoleProtectionPolicy",
            dest="WriteHoleProtectionPolicy",
            help="""Optionally include to choose the WriteHoleProtectionPolicy """
            """this is applicable only for VROC, You can send either Yes or No as values""",
            default=None,
        )
        v_parser.add_argument(
            "--sparedrives",
            dest="sparedrives",
            help="""Optionally include to set the spare drives by the """
            """physical drive's location. (usable in custom creation only)""",
            action="append",
            default=None,
        )
        v_parser.add_argument(
            "--capacitybytes",
            dest="capacitybytes",
            help="""Optionally include to set the capacity of the drive in """
            """bytes. (usable in custom creation only, use -1 for max """
            """size)""",
            action="append",
            default=None,
        )
        v_parser.add_argument(
            "--controller",
            dest="controller",
            help="Use this flag to select the corresponding controller "
            "using either the slot number or index.\nexample: --controller=Slot 0 OR "
            "--controller=1",
            default=None,
            required=True,
        )

        self.cmdbase.add_login_arguments_group(v_parser)
        # self.options_argument_group(v_parser)



"""Module containing functional tests for the 'applydcpmmconfig' command"""
from __future__ import absolute_import

import sys

from rdmc_helper import ReturnCodes


class TestApplyPmemConfigCommand(object):
    """Class for functional tests for the applydcpmmconfig command"""

    @staticmethod
    def test_show_invalid_flag(ilorest):
        """Test function call with invalid flag"""
        results = ilorest.run(['applydcpmmconfig', '--bad-flag'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_show_help_flag(ilorest):
        """Test function call with help flag"""
        results = ilorest.run(['applydcpmmconfig', '-h'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_no_flag(ilorest):
        """Test for the case when `applydcpmmconfig` command is run without any flag"""
        results = ilorest.run(['applydcpmmconfig'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_force_flag_1(ilorest):
        """Test for the case when `applydcpmmconfig` command is run with force flag"""
        results = ilorest.run(['applydcpmmconfig', '-f'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_force_flag_2(ilorest):
        """Test for the case when `applydcpmmconfig` command is run with force flag"""
        results = ilorest.run(['applydcpmmconfig', '-L', '-f'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    # Cannot test for the positive case - mock server not accepting delete tasks.
    @staticmethod
    def test_with_force_flag_3(ilorest, dl360_large):
        """Test for the case when `applydcpmmconfig` command is run with force flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['applydcpmmconfig', '-C', 'MemoryMode', '-f'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.NO_CHANGES_MADE_OR_FOUND

    @staticmethod
    def test_with_config_flag_1(ilorest, dl360_large):
        """Test for the case when `applydcpmmconfig` command is run with config flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['applydcpmmconfig', '-C'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_config_flag_2(ilorest, dl360_large):
        """Test for the case when `applydcpmmconfig` command is run with config flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['applydcpmmconfig', '-C', 'xyz'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_config_flag_3(ilorest, dl360_large):
        """Test for the case when `applydcpmmconfig` command is run with config flag"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['applydcpmmconfig', '-C', 'AppDirectInterleaved'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.NO_CHANGES_MADE_OR_FOUND

    @staticmethod
    def test_with_list_flag(ilorest):
        """Test for the case when `applydcpmmconfig` command is run with list flag"""
        results = ilorest.run(['applydcpmmconfig', '-L'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.SUCCESS

    @staticmethod
    def test_with_extra_attributes_1(ilorest):
        """Test for the case when `applydcpmmconfig` command is run with extraneous attributes"""
        results = ilorest.run(['applydcpmmconfig', '-L', '1qazxcsd'])
        sys.stdout.write("results.stdout: ")
        sys.stdout.write(results.stdout)
        sys.stdout.write("results.stderr: ")
        sys.stdout.write(results.stderr)
        assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR

    @staticmethod
    def test_with_extra_attributes_2(ilorest, dl360_large):
        """Test for the case when `applydcpmmconfig` command is run with extraneous attributes"""
        url, username, password = dl360_large.config
        with ilorest.session(url, username, password):
            results = ilorest.run(['applydcpmmconfig', '-C', 'AppDirectInterleaved', '1qazxcsd'])
            sys.stdout.write("results.stdout: ")
            sys.stdout.write(results.stdout)
            sys.stdout.write("results.stderr: ")
            sys.stdout.write(results.stderr)
            assert results.retcode == ReturnCodes.INVALID_COMMAND_LINE_ERROR



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" SetPassword Command for rdmc """

try:
    from rdmc_helper import (
        Encryption,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
        UnableToDecodeError,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        Encryption,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
        UnableToDecodeError,
    )


class SetPasswordCommand:
    """Set password class command"""

    def __init__(self):
        self.ident = {
            "name": "setpassword",
            "usage": None,
            "description": "Sets the admin password and power-on password\n"
            "setpassword --newpassword <NEW_PASSWORD> --currentpassword <OLD_PASSWORD> [OPTIONS]\n\n\t"
            "Setting the admin password with no previous password set."
            "\n\texample: setpassword --newpassword testnew --currentpassword None\n\n\tSetting the admin "
            "password back to nothing.\n\texample: setpassword --newpassword None --currentpassword testnew "
            "\n\n\tSetting the power on password.\n\texample: setpassword"
            " --newpassword testnew --currentpassword None --poweron\n\tNote: "
            "if it is empty password, send None as above.",
            "summary": "Sets the admin password and power-on password",
            "aliases": [],
            "auxcommands": [
                "LoginCommand",
                "SetCommand",
                "SelectCommand",
                "CommitCommand",
                "RebootCommand",
                "LogoutCommand",
            ],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main set password worker function

        :param line: string of arguments passed in
        :type line: str.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.setpasswordvalidation(options)

        # if not args:
        #    self.rdmc.ui.printer('Please input the current password.\n')
        #    tempoldpass = getpass.getpass()

        #   if tempoldpass and tempoldpass != '\r':
        #        tempoldpass = tempoldpass
        #    else:
        #        tempoldpass = '""'

        #    self.rdmc.ui.printer('Please input the new password.\n')
        #    tempnewpass = getpass.getpass()

        #    if tempnewpass and tempnewpass != '\r':
        #        tempnewpass = tempnewpass
        #    else:
        #        tempnewpass = '""'
        #    args.extend([tempnewpass, tempoldpass])

        # if len(args) < 2:
        #    raise InvalidCommandLineError("Please pass both new password and old password.")

        args = list()
        args.append(options.newpassword)
        args.append(options.currentpassword)
        count = 0
        for arg in args:
            if arg:
                if arg.lower() == "none" or arg.lower() == "null" or arg is None:
                    args[count] = ""
                elif len(arg) > 2:
                    if ('"' in arg[0] and '"' in arg[-1]) or ("'" in arg[0] and "'" in arg[-1]):
                        args[count] = arg[1:-1]
                elif len(arg) == 2:
                    if (arg[0] == '"' and arg[1] == '"') or (arg[0] == "'" and arg[1] == "'"):
                        args[count] = ""
            count += 1

        if options.encode:
            _args = []
            for arg in args:
                try:
                    arg = Encryption.decode_credentials(arg)
                    if isinstance(arg, bytes):
                        arg = arg.decode("utf-8")
                    _args.append(arg)
                except UnableToDecodeError:
                    _args.append(arg)
            args = _args
        if self.rdmc.app.typepath.defs.isgen10:
            bodydict = self.rdmc.app.get_handler(self.rdmc.app.typepath.defs.biospath, service=True, silent=True).dict

            for item in bodydict["Actions"]:
                if "ChangePassword" in item:
                    path = bodydict["Actions"][item]["target"]
                    break

            if options.poweron:
                body = {
                    "PasswordName": "User",
                    "OldPassword": args[1],
                    "NewPassword": args[0],
                }
            else:
                body = {
                    "PasswordName": "Administrator",
                    "OldPassword": args[1],
                    "NewPassword": args[0],
                }

            self.rdmc.app.post_handler(path, body)
        else:
            if options.poweron:
                self.auxcommands["select"].run("HpBios.")
                self.auxcommands["set"].run("PowerOnPassword=%s OldPowerOnPassword=%s" % (args[0], args[1]))
                self.auxcommands["commit"].run("")
            else:
                self.auxcommands["select"].run("HpBios.")
                self.auxcommands["set"].run("AdminPassword=%s OldAdminPassword=%s" % (args[0], args[1]))
                self.auxcommands["commit"].run("")
                self.rdmc.ui.printer(
                    "\nThe session will now be terminated.\n"
                    " login again with updated credentials in order to continue.\n"
                )
                self.auxcommands["logout"].run("")

        if options:
            if options.reboot:
                self.auxcommands["reboot"].run(options.reboot)

        self.cmdbase.logout_routine(self, options)
        return ReturnCodes.SUCCESS

    def setpasswordvalidation(self, options):
        """Results method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--currentpassword",
            dest="currentpassword",
            help="Use this flag to provide current password.",
            required=True,
        )

        customparser.add_argument(
            "--newpassword",
            dest="newpassword",
            help="Use this flag to provide new password.",
            required=True,
        )

        customparser.add_argument(
            "--reboot",
            dest="reboot",
            help="Use this flag to perform a reboot command function after "
            "completion of operations. 'REBOOT' is a replaceable parameter "
            "that can have multiple values. For help with parameters and "
            "descriptions regarding the reboot flag, run help reboot.",
            default=None,
        )
        customparser.add_argument(
            "--poweron",
            dest="poweron",
            action="store_true",
            help="""Use this flag to set power on password instead""",
            default=None,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Select Command for RDMC """

from redfish.ris import NothingSelectedError

try:
    from rdmc_helper import LOGGER, InvalidCommandLineErrorOPTS, ReturnCodes
except ImportError:
    from ilorest.rdmc_helper import LOGGER, InvalidCommandLineErrorOPTS, ReturnCodes


class SelectCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "select",
            "usage": None,
            "description": "Selects the Redfish/HpRest type to be used.\nIn order to "
            "remove the need of including the version while selecting you"
            " can simply enter the type name until the first period\n\t"
            "example: select HpBios.\n"
            "Run without an argument to "
            "display the currently selected type\n\texample: select",
            "summary": "Selects the object type to be used.",
            "aliases": ["sel"],
            "auxcommands": ["LoginCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def selectfunction(self, line):
        """Main select worker function

        :param line: command line input
        :type line: string.
        """

        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.selectvalidation(options)

        if args:
            if options.ref:
                LOGGER.warning("Patches from current selection will be cleared.")
            selector = args[0]
            selections = self.rdmc.app.select(selector=selector, path_refresh=options.ref)

            if "storage" in selector.lower():
                storage_url = "/redfish/v1/Systems/1/Storage/"
                list_storageid = self.rdmc.app.get_handler(storage_url, silent=True, service=True).dict["Members"]
                for val in list_storageid:
                    storageid_fromurl = val["@odata.id"]
                    ctr = self.rdmc.app.get_handler(storageid_fromurl, silent=True, service=True).dict
                    if "Controllers" in ctr:
                        ctr_url = ctr["Controllers"]["@odata.id"]
                        self.rdmc.app.get_handler(ctr_url, silent=True, service=True).dict
            if self.rdmc.opts.verbose and selections:
                templist = list()
                self.rdmc.ui.printer("Selected option(s): ")

                for item in selections:
                    if item.type not in templist:
                        templist.append(item.type)

                self.rdmc.ui.printer("%s\n" % ", ".join(map(str, templist)))

        else:
            selector = self.rdmc.app.selector

            if selector:
                sellist = [sel for sel in self.rdmc.app.monolith.typesadded if selector.lower() in sel.lower()]
                self.rdmc.ui.printer("Current selection: ")
                self.rdmc.ui.printer("%s\n" % ", ".join(map(str, sellist)))
            else:
                raise NothingSelectedError

        self.cmdbase.logout_routine(self, options)

    def selectvalidation(self, options):
        """Select data validation function

        :param options: command line options
        :type options: list.
        """

        self.cmdbase.login_select_validation(self, options)

    def run(self, line, help_disp=False):
        """Wrapper function for main select function

        :param line: entered command line
        :type line: list.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        self.selectfunction(line)

        # Return code
        return ReturnCodes.SUCCESS

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--refresh",
            dest="ref",
            action="store_true",
            help="Optionally reload the data of selected type and clear " "patches from current selection.",
            default=False,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""This module contains common helper functions used by several pmem commands"""


class PmemHelpers(object):
    """
    Class containing common helper functions used by several pmem commands
    """

    @staticmethod
    def py3_round(number, precision):
        """
        Rounds numbers in accordance with the Python 3 round specification
        :param number: number to be rounded
        :type number: floating point number
        :param precision: Number of decimal places the number should be rounded off to
        :type precision: integer
        :return: rounded off number
        """
        if abs(round(number) - number) == 0.5:
            return 2.0 * round(number / 2.0, precision)
        return round(number, precision)

    @staticmethod
    def parse_dimm_id(dimm_id_list):
        """
        Converts DIMM IDs from the 'X@Y' format
        to the 'PROC X DIMM Y' format
        :param dimm_id_list: DIMM IDs in the 'X@Y' format
        :type dimm_id_list: list
        :return: list of DIMM IDs in the 'PROC X DIMM Y' format
        """
        parsed_list = list()
        for dimm_id in dimm_id_list:
            temp = dimm_id.split("@")
            parsed_list.append("PROC " + temp[0] + " DIMM " + temp[1])
        return parsed_list

    @staticmethod
    def get_pmem_members(memory_members):
        """
        Filters persistent memory members from memory resources
        :param memory_members: members of memory collection resource
        :type memory_members: list of members
        :returns: list of persistent memory members if found else empty list
        """
        base_module_type = "PMM"
        pmem_members = list()
        pmem_dimm_id = set()
        for member in memory_members:
            memory_type = member.get("Oem").get("Hpe").get("BaseModuleType")
            if memory_type == base_module_type:
                pmem_members.append(member)
                pmem_dimm_id.add(member.get("DeviceLocator"))
        return pmem_members, pmem_dimm_id

    @staticmethod
    def get_non_aep_members(memory_members):
        """
        Filters dram memory members from memory resources
        :param memory_members: members of memory collection resource
        :type memory_members: list of members
        :returns: list of dram memory members if found else empty list
        """
        base_module_type = "PMM"
        dram_members = list()
        dram_dimm_id = set()
        for member in memory_members:
            memory_type = member.get("Oem").get("Hpe").get("BaseModuleType")
            if memory_type != base_module_type:
                dram_members.append(member)
                dram_dimm_id.add(member.get("DeviceLocator"))
        return dram_members, dram_dimm_id

    @staticmethod
    def json_to_text(dictionary):
        """
        Converts json to string format
        :param dictionary: json to be converted
        :return: list containing the string
        """
        output = ""
        for key, value in dictionary.items():
            item = "\n" + key + ":" + str(value)
            output += item
        return [output]

    @staticmethod
    def location_format_converter(location_list):
        """
        Converts location format from 'PROC X DIMM Y' to 'X@Y'
        :param location_list: list of locations of the format 'PROC X DIMM Y'
        :type location_list: list of strings
        :returns: string of locations in the 'X@Y' format (comma separated)
        """
        converted_str = ""
        for location in location_list:
            temp = location.split(" ")
            converted_str += temp[1] + "@" + temp[3]
            if location is not location_list[-1]:
                converted_str += ", "
        return converted_str, ("PROC " + converted_str[0])

    @staticmethod
    def compare_id(id1, id2):
        """
        Compares two ids
        :param id1: first id to be compared
        :param id2: second id to be compared
        :return: True if ids are same else False
        """
        if id1[-1] == "/":
            id1 = id1[:-1]
        if id2[-1] == "/":
            id2 = id2[:-1]
        return id1 == id2



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Command to show recommended configurations"""
from __future__ import absolute_import, division

from collections import OrderedDict

try:
    from rdmc_helper import (
        LOGGER,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoContentsFoundForOperationError,
        LOGGER,
    )

from .lib.DisplayHelpers import DisplayHelpers, OutputFormats
from .lib.Mapper import Mapper
from .lib.MapperRenderers import MappingTable
from .lib.PmemHelpers import PmemHelpers
from .lib.RestHelpers import RestHelpers


class ShowRecommendedConfigCommand:
    """Show recommended configurations"""

    def __init__(self):
        self.ident = {
            "name": "showrecommendedpmmconfig",
            "usage": None,
            "description": "Show recommended configurations\n" "\texample: showrecommendedpmmconfig",
            "summary": "Show Recommended Configuration",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()
        self._display_helpers = DisplayHelpers()
        self._mapper = Mapper()
        self._pmem_helpers = PmemHelpers()

    def definearguments(self, customparser):
        """
        Wrapper function for new command main function
        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-j",
            "--json",
            action="store_true",
            dest="json",
            help="Optionally include this flag to change the output to JSON format.",
            default=False,
        )

    def run(self, line, help_disp=False):
        """
        Wrapper function for new command main function
        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        LOGGER.info("Show Recommended Configuration: %s", self.ident["name"])
        try:
            (_, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineError("Failed to parse options")
        if args:
            raise InvalidCommandLineError("Chosen command doesn't expect additional arguments")
        # Raise exception if server is in POST
        if RestHelpers(rdmcObject=self.rdmc).in_post():
            raise NoContentsFoundForOperationError(
                "Unable to retrieve resources - " "server might be in POST or powered off"
            )
        self.show_recommended_config()

        return ReturnCodes.SUCCESS

    def show_recommended_config(self):
        """
        Show recommended pmm configuration
        """
        members = RestHelpers(rdmcObject=self.rdmc).retrieve_memory_resources().get("Members")

        if not members:
            raise NoContentsFoundForOperationError("Failed to retrieve memory resources")

        # Retreving dimms
        pmem_members = self._pmem_helpers.get_pmem_members(members)[0]

        if not pmem_members:
            raise NoContentsFoundForOperationError("No Persistent Memory Modules found")

        # Retreving dram dimms
        dram_members = self._pmem_helpers.get_non_aep_members(members)[0]

        if not dram_members:
            raise NoContentsFoundForOperationError("No DRAM DIMMs found")

        # retrieving Total Capacity of PMEM dimms
        attr = self._mapper.get_single_attribute(pmem_members, "TotalCapacity", MappingTable.summary.value, True)
        pmem_size = attr.get("TotalCapacity", {}).get("Value", 0)

        # retrieving Total Capacity of DRAM dimms
        dram_size = self._mapper.get_single_attribute(dram_members, "TotalCapacity", MappingTable.summary.value, True)
        dram_size = dram_size.get("TotalCapacity", {}).get("Value", 0)

        display_output = list()
        recommended_config = list()

        # Add AppDirect Mode
        temp_dict = OrderedDict()
        temp_dict["MemoryModeTotalSize"] = 0
        temp_dict["PmemTotalSize"] = pmem_size
        temp_dict["CacheRatio"] = None
        recommended_config.append(temp_dict)

        # Add Memory Mode
        temp_dict = OrderedDict()
        temp_dict["MemoryModeTotalSize"] = pmem_size
        temp_dict["PmemTotalSize"] = 0
        temp_dict["CacheRatio"] = pmem_size / dram_size
        recommended_config.append(temp_dict)

        # Add Mixed Mode (BPS doesn't support it)
        if "Gen10 Plus" not in RestHelpers(rdmcObject=self.rdmc).retrieve_model(self.rdmc)["Model"]:
            stepsize = 32
            appdirect_size = 0
            step = 1
            while appdirect_size < pmem_size:
                appdirect_size = step * stepsize * len(pmem_members)
                memorymode_size = pmem_size - appdirect_size
                cache_ratio = memorymode_size / dram_size
                if 2 <= cache_ratio <= 16:
                    temp_dict = OrderedDict()
                    temp_dict["MemoryModeTotalSize"] = memorymode_size
                    temp_dict["PmemTotalSize"] = appdirect_size
                    temp_dict["CacheRatio"] = cache_ratio
                    recommended_config.append(temp_dict)
                step += 1

        # Sorting based on MemoryModeTotalSize
        recommended_config = sorted(recommended_config, key=lambda x: x["MemoryModeTotalSize"])

        # Adding units and formating cache ratio
        for output in recommended_config:
            output["MemoryModeTotalSize"] = "%d GB" % output["MemoryModeTotalSize"]
            output["PmemTotalSize"] = "%d GB" % output["PmemTotalSize"]
            if output["CacheRatio"] is None:
                output["CacheRatio"] = "N/A"
            else:
                output["CacheRatio"] = "1:%.1f" % output["CacheRatio"]
            display_output.append(self._pmem_helpers.json_to_text(output)[0])

        # Display output in table format
        self._display_helpers.display_data(display_output, OutputFormats.table)



###
# Copyright 2016 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Monolith Command for rdmc """

import json

try:
    from rdmc_helper import UI, InvalidCommandLineErrorOPTS, ReturnCodes
except ImportError:
    from ilorest.rdmc_helper import UI, InvalidCommandLineErrorOPTS, ReturnCodes
from redfish.rest.containers import JSONEncoder
from redfish.ris import UndefinedClientError


class MonolithCommand:
    """Monolith class command"""

    def __init__(self):
        self.ident = {
            "name": "monolith",
            "usage": None,
            "description": "Displays entire cached data structure available",
            "summary": "displays entire cached data structure available",
            "aliases": ["mono"],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main monolith worker function

        :param line: string of arguments passed in
        :type line: str.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if hasattr(self.rdmc.app, "config") and self.rdmc.app.config._ac__format.lower() == "json":
            options.json = True

        if self.rdmc.app.current_client:
            results = self.rdmc.app.monolith.capture(redmono=options.redmono)
        else:
            raise UndefinedClientError()

        if options.filename:
            with open(options.filename[0], "w") as monolith:
                if options.json:
                    json.dump(results, monolith, indent=2, cls=JSONEncoder)
                else:
                    monolith.write(str(results))

        elif options.json:
            UI().print_out_json(results)  # .reduce())
        else:
            UI().print_out_human_readable(results)  # .reduce())

        # Return code
        return ReturnCodes.SUCCESS

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        customparser.add_argument(
            "-f",
            "--filename",
            dest="filename",
            help="Use this flag if you wish to save the monolith" " into a file with the given filename.",
            action="append",
            default=None,
        )
        customparser.add_argument(
            "-r",
            "--reduced",
            dest="redmono",
            action="store_true",
            help="Use this flag if you wish to save the reduced monolith" " into a file with the given filename.",
            default=False,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
""" BiosDefaultsCommand for rdmc """

from argparse import RawDescriptionHelpFormatter

try:
    from rdmc_helper import (
        LOGGER,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidPasswordLengthError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        LOGGER,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidPasswordLengthError,
        ReturnCodes,
    )


class ServerConfigurationLockCommand:
    """Set BIOS settings back to default for the server that is currently
    logged in"""

    def __init__(self):
        self.ident = {
            "name": "serverconfiglock",
            "usage": None,
            "description": "To to perform validation checks on raw JSON data "
            "server.\n\texample: serverconfiglock\n\t"
            "example: serverconfiglock enable --serverconfiglockpassword=AWS@123456789 "
            "--serverconfiglockexcludefwrevs=True\t serverconfiglockexcludefwrevs set True or False\n\t"
            "example: serverconfiglock disable --serverconfiglockpassword=AWS@1234 "
            "--serverconfiglockdisable=True\t serverconfiglockdisable set True or False\n\t"
            "example: serverconfiglock display",
            "summary": "The BIOS feature “Server Configuration Lock” supports certain parameters,"
            "including a password. This password has a 16 to 31 character limit. "
            "“Server Configuration Lock” is not one of the special commands in iLO REST.",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main BIOS defaults worker function"""
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")
        bios_path = self.rdmc.app.typepath.defs.biospath
        resp = self.rdmc.app.get_handler(bios_path, silent=True, service=True).dict
        match_url = resp["Oem"]["Hpe"]["Links"]["ServerConfigLock"]["@odata.id"]
        if "oem" in match_url:
            scs_path = bios_path + "oem/hpe/serverconfiglock/settings/"
        else:
            scs_path = bios_path + "serverconfiglock/settings/"
        self.serverconfiglockvalidation(options)
        if options.command:
            if options.command.lower() == "enable":
                self.enable_scs(options, scs_path)
            elif options.command.lower() == "disable":
                self.disable_scs(options, scs_path)
            elif options.command == "display":
                self.display_scs(scs_path)
            else:
                raise InvalidCommandLineError("%s is not a valid option for this " "command." % str(options.command))
        else:
            raise InvalidCommandLineError(
                "Please provide either enable, disable and display subcommand. "
                "For help or usage related information, use -h or --help"
            )
        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def enable_scs(self, options, scs_path):
        """
        Enable/Disable the SCL config function
        """
        if options.serverconfiglockpassword:
            ser_cf_password = options.serverconfiglockpassword
        self.serverconfiglockpassword_validation(options)
        if options.serverconfiglockexcludefwrevs.lower() == "true":
            serverconfrevs = "True"
        elif options.serverconfiglockexcludefwrevs.lower() == "false":
            serverconfrevs = "False"
        else:
            raise InvalidCommandLineError("ServerConfigLockExcludeFwRevs value invalid set only true or false")
        body = {"ServerConfigLockPassword": ser_cf_password, "ServerConfigLockExcludeFwRevs": serverconfrevs}
        try:
            self.rdmc.ui.printer("payload: %s \n" % body)
            LOGGER.info("payload", body)
            resp = self.rdmc.app.patch_handler(scs_path, body)
            if resp.status == 200:
                self.rdmc.ui.printer("SCS enabled successfully\n")
                LOGGER.info("SCS enabled successfully")
                self.rdmc.ui.printer("Help: System Reboot is required after changing the serverconfiglock settings \n")
                return ReturnCodes.SUCCESS
        except IncompatibleiLOVersionError:
            self.rdmc.ui.error("iLO FW version on this server doesnt support SCS")
            LOGGER.error("iLO FW version on this server doesnt support SCS")
            return ReturnCodes.INCOMPATIBLE_ILO_VERSION_ERROR

    def disable_scs(self, options, scs_path):
        """
        Disable the SCL config function
        """
        if options.serverconfiglockpassword:
            ser_cf_password = options.serverconfiglockpassword
        self.serverconfiglockpassword_validation(options)
        if options.serverconfiglockdisable.lower() == "true":
            sclockdisable = "True"
        elif options.serverconfiglockdisable.lower() == "false":
            sclockdisable = "False"
        else:
            raise InvalidCommandLineError("ServerConfigLockDisable value invalid set only true or false")
        body = {"ServerConfigLockPassword": ser_cf_password, "ServerConfigLockDisable": sclockdisable}
        try:
            self.rdmc.ui.printer("payload: %s \n" % body)
            LOGGER.info("payload", body)
            resp = self.rdmc.app.patch_handler(scs_path, body)
            if resp.status == 200:
                self.rdmc.ui.printer("SCS Disabled successfully\n")
                LOGGER.info("SCS Disabled successfully")
                self.rdmc.ui.printer("Help: System Reboot is required after changing the serverconfiglock settings \n")
                return ReturnCodes.SUCCESS
        except IncompatibleiLOVersionError:
            self.rdmc.ui.error("iLO FW version on this server doesnt support SCS")
            LOGGER.error("iLO FW version on this server doesnt support SCS")
            return ReturnCodes.INCOMPATIBLE_ILO_VERSION_ERROR

    def display_scs(self, scs_path):
        """
        Display SCS details
        """
        ilo_ver = self.rdmc.app.getiloversion()
        if ilo_ver >= 5.290 or ilo_ver < 6.000:
            results = self.rdmc.app.get_handler(scs_path, silent=True)
            if results.status == 200:
                self.rdmc.ui.printer("Server Configuration Lock setting details ...\n")
                LOGGER.info("Server Configuration Lock setting details ...")
                results = results.dict
                self.print_scs_info(results)
        else:
            self.rdmc.ui.printer("For iLO6, this command is not yet supported.\n")
            LOGGER.error("For iLO6, this command is not yet supported.")

    def print_scs_info(self, results):
        """
        Prints the SCL: info
        """
        for key, value in results.items():
            if "@odata" not in key:
                if type(value) is dict:
                    self.print_cert_info(value)
                else:
                    self.rdmc.ui.printer(key + ":" + str(value) + "\n")

    def serverconfiglockpassword_validation(self, options):
        """
        Password validation for ServerConfigLockPassword method
        """
        import re

        if options.serverconfiglockpassword:
            password = options.serverconfiglockpassword
        if len(password) < 16 or len(password) > 31:
            LOGGER.info("Length of password requires min of 16 and max of 31 character " "excluding '/' character.\n")
            raise InvalidPasswordLengthError(
                "Length of password requires min of 16 and max of 31 character " "excluding '/' character.\n"
            )
        elif re.search("[0-9]", password) is None:
            print("Make sure your password has a number in it\n")
        elif re.search("[A-Z]", password) is None:
            print("Make sure your password has a capital letter in it\n")

    def serverconfiglockvalidation(self, options):
        """new command method validation function"""
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        subcommand_parser = customparser.add_subparsers(dest="command")
        enable_help = "To enable ServerConfigurationLock \n"

        enable_parser = subcommand_parser.add_parser(
            "enable",
            help=enable_help,
            description=enable_help + "\n\tExample:\n\tserverconfiglock enable ",
            formatter_class=RawDescriptionHelpFormatter,
        )
        enable_parser.add_argument(
            "--serverconfiglockpassword",
            dest="serverconfiglockpassword",
            help="Set a serverconfiglockpassword",
            required=True,
            type=str,
            default=None,
        )
        enable_parser.add_argument(
            "--serverconfiglockexcludefwrevs",
            dest="serverconfiglockexcludefwrevs",
            help="to set/enable serverconfiglockexcludefwrevs",
            default=False,
            required=True,
        )
        self.cmdbase.add_login_arguments_group(enable_parser)

        status_help = "To check the ServerConfigurationLock display\n"
        status_parser = subcommand_parser.add_parser(
            "display",
            help=status_help,
            description=status_help + "\n\tExample:\n\tserverconfiglock display or " "\n\tserverconfiglock status -j",
            formatter_class=RawDescriptionHelpFormatter,
        )

        self.cmdbase.add_login_arguments_group(status_parser)

        disable_help = "To disable the ServerConfigurationLock\n"
        disable_parser = subcommand_parser.add_parser(
            "disable",
            help=disable_help,
            description=disable_help + "\n\tExample:\n\tserverconfiglock disable",
            formatter_class=RawDescriptionHelpFormatter,
        )
        disable_parser.add_argument(
            "--serverconfiglockpassword",
            dest="serverconfiglockpassword",
            help="Server Configuration Lock password and to digitally fingerprint the system to enable Server "
            "Configuration Lock.",
            required=True,
            type=str,
            default=None,
        )
        disable_parser.add_argument(
            "--serverconfiglockdisable",
            dest="serverconfiglockdisable",
            help="Select this option to disable Server Configuration Lock.",
            required=True,
            default=False,
        )
        self.cmdbase.add_login_arguments_group(disable_parser)



###
# Copyright 2021-2022 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
""" computeopsmanagement Command for rdmc """

import time
from argparse import RawDescriptionHelpFormatter

try:
    from rdmc_helper import (
        CloudConnectFailedError,
        CloudConnectTimeoutError,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoCurrentSessionEstablished,
        ProxyConfigFailedError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        CloudConnectFailedError,
        CloudConnectTimeoutError,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoCurrentSessionEstablished,
        ProxyConfigFailedError,
        ReturnCodes,
    )

from redfish.ris.ris import SessionExpired


class ComputeOpsManagementCommand:
    """Main new command template class"""

    def __init__(self):
        self.ident = {
            "name": "computeopsmanagement",
            "usage": "computeopsmanagement\n\n",
            "description": "Run to enable your servers to be discovered, "
            "monitored and managed through ComputeOpsManagement\n\t"
            "Example:\n\tcomputeopsmanagement connect or \n\t"
            "computeopsmanagement connect --activationkey <ACTIVATION KEY> or \n\t"
            "computeopsmanagement connect --activationkey <ACTIVATION KEY> --proxy http://proxy.abc.com:8080 or \n\t"
            "computeopsmanagement disconnect or \n\t"
            "computeopsmanagement status or \n\t"
            "computeopsmanagement status -j\n",
            "summary": "Enables the server to be discovered, monitored and managed through ComputeOpsManagement",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def proxy_config(self, proxy_server):
        """Main cloudconnect worker function

        :param proxy_server: proxy
        :type proxy_server: str.
        """
        if proxy_server != "None":
            try:
                body = dict()
                body["Oem"] = {}
                body["Oem"]["Hpe"] = {}
                body["Oem"]["Hpe"]["WebProxyConfiguration"] = {}
                proxy_body = body["Oem"]["Hpe"]["WebProxyConfiguration"]
                proxy_body["ProxyServer"] = None
                proxy_body["ProxyUserName"] = None
                proxy_body["ProxyPassword"] = None
                if "https" in proxy_server:
                    proxy_body["ProxyPort"] = 443
                else:
                    proxy_body["ProxyPort"] = 80
                if "@" in proxy_server:
                    proxy = proxy_server.split("@")
                    proxy_usr_pass = proxy[0]
                    proxy_srv_port = proxy[1]
                    if "//" in proxy_usr_pass:
                        proxy_usr_pass = proxy_usr_pass.split("//")[1]
                    if ":" in proxy_srv_port:
                        proxy = proxy_srv_port.split(":")
                        proxy_body["ProxyServer"] = proxy[0]
                        proxy_body["ProxyPort"] = int(proxy[1])
                    else:
                        proxy_body["ProxyServer"] = proxy_srv_port
                    if ":" in proxy_usr_pass:
                        proxy = proxy_usr_pass.split(":")
                        proxy_body["ProxyPassword"] = proxy[1]
                        proxy_body["ProxyUserName"] = proxy[0]
                    else:
                        proxy_body["ProxyUserName"] = proxy_usr_pass
                else:
                    proxy_srv_port = proxy_server
                    if "//" in proxy_srv_port:
                        proxy_srv_port = proxy_srv_port.split("//")[1]
                    if ":" in proxy_srv_port:
                        proxy = proxy_srv_port.split(":")
                        proxy_body["ProxyServer"] = proxy[0]
                        proxy_body["ProxyPort"] = int(proxy[1])
                    else:
                        proxy_body["ProxyServer"] = proxy_srv_port

                path = self.rdmc.app.getidbytype("NetworkProtocol.")

                if path and body:
                    self.rdmc.ui.printer("Setting Proxy configuration...\n", verbose_override=True)
                    self.rdmc.app.patch_handler(path[0], body, service=False, silent=True)
            except:
                raise ProxyConfigFailedError("Setting Proxy Server Configuration Failed.\n")
        else:
            try:
                body = dict()
                body["Oem"] = {}
                body["Oem"]["Hpe"] = {}
                body["Oem"]["Hpe"]["WebProxyConfiguration"] = {}
                proxy_body = body["Oem"]["Hpe"]["WebProxyConfiguration"]
                proxy_body["ProxyServer"] = ""
                proxy_body["ProxyPort"] = None
                proxy_body["ProxyUserName"] = ""
                proxy_body["ProxyPassword"] = None
                path = self.rdmc.app.getidbytype("NetworkProtocol.")

                if path and body:
                    self.rdmc.ui.printer("Clearing Proxy configuration...\n", verbose_override=True)
                    self.rdmc.app.patch_handler(path[0], body, service=False, silent=True)
            except:
                raise ProxyConfigFailedError("Clearing Proxy Server Configuration Failed.\n")

    def get_cloud_status(self):
        path = self.rdmc.app.typepath.defs.managerpath
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)
        if resp.status != 200:
            raise SessionExpired("Invalid session. Please logout and log back in or include credentials.")
        status = resp.dict["Oem"]["Hpe"]["CloudConnect"]["CloudConnectStatus"]
        return status

    def connect_cloud(self, activationkey=None):
        """cloud connect function

        :param activationkey: activation key
        :type activationkey: str.
        """
        status = self.get_cloud_status()
        if status == "Connected":
            self.rdmc.ui.printer("Warning: ComputeOpsManagement is already connected.\n")
            return ReturnCodes.SUCCESS
        body = dict()
        # Temporary
        # body['CloudActivateURL'] = "https://qa-devices.rugby.hpeserver.management/inventory/compute-provision"
        if activationkey:
            body["ActivationKey"] = activationkey
        else:
            body = {}
        path = self.rdmc.app.typepath.defs.managerpath + "Actions" + self.rdmc.app.typepath.defs.oempath
        path = path + "/HpeiLO.EnableCloudConnect"
        try:
            if path:
                self.rdmc.ui.printer("Connecting to ComputeOpsManagement...", verbose_override=True)
                self.rdmc.app.post_handler(path, body, service=False, silent=True)
        except:
            raise CloudConnectFailedError("ComputeOpsManagement connection Failed.\n")
        start_time = time.time()
        allowed_seconds = 120
        time_increment = 5
        i = 1
        while True:
            # time.sleep(time_increment * i)
            time.sleep(time_increment)
            current_time = time.time()
            elapsed_time = current_time - start_time
            if elapsed_time > allowed_seconds:
                self.rdmc.ui.printer("\n")
                raise CloudConnectTimeoutError(
                    "ComputeOpsManagement connection timed out, Please check the "
                    "activation key and network or proxy settings and try again.\n"
                )
            else:
                status = self.get_cloud_status()
                # self.rdmc.ui.printer("ComputeOpsManagement connection status is %s.\n" % status)
                if status == "Connected":
                    # Check again after 10 seconds before breaking the loop
                    time.sleep(10)
                    status = self.get_cloud_status()
                    if status == "Connected":
                        self.rdmc.ui.printer("\n")
                        self.rdmc.ui.printer("ComputeOpsManagement connection is successful.\n")
                        break
                # If connection has failed while checking waiting for success message
                if status == "ConnectionFailed":
                    self.rdmc.ui.printer("\n")
                    raise CloudConnectFailedError(
                        "ComputeOpsManagement connection Failed. Please check the "
                        "activation key and network or proxy settings and try again.\n"
                    )
                else:
                    self.rdmc.ui.printer("..")
                    i = i + 1

    def disconnect_cloud(self):
        """cloud disconnect function"""
        cloud_status = self.get_cloud_status()
        if cloud_status == "Connected" or cloud_status == "ConnectionFailed":
            path = self.rdmc.app.typepath.defs.managerpath + "Actions" + self.rdmc.app.typepath.defs.oempath
            path = path + "/HpeiLO.DisableCloudConnect"
            body = dict()
            try:
                if path:
                    self.rdmc.ui.printer("Disconnecting ComputeOpsManagement...\n", verbose_override=True)
                    self.rdmc.app.post_handler(path, body)
                    time.sleep(10)
                    cloud_status = self.get_cloud_status()
                    if cloud_status == "NotEnabled":
                        self.rdmc.ui.printer("The operation completed successfully.\n")
            except:
                raise CloudConnectFailedError("ComputeOpsManagement is not disconnected.\n")
        else:
            self.rdmc.ui.printer(
                "Warning: ComputeOpsManagement is not at all connected.\n",
                verbose_override=True,
            )

    def cloud_status(self, json=False):
        """cloud connect function

        :param json: json
        :type json: bool
        """
        path = self.rdmc.app.typepath.defs.managerpath
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)
        if resp.status != 200:
            raise SessionExpired("Invalid session. Please logout and log back in or include credentials.")
        cloud_info = resp.dict["Oem"]["Hpe"]["CloudConnect"]
        output = "------------------------------------------------\n"
        output += "ComputeOpsManagement connection status\n"
        output += "------------------------------------------------\n"
        output += "ComputeOpsManagement Status : %s\n" % (cloud_info["CloudConnectStatus"])
        if cloud_info["CloudConnectStatus"] != "NotEnabled":
            if "CloudActivateURL" in cloud_info:
                output += "CloudActivateURL : %s\n" % (cloud_info["CloudActivateURL"])
            if "ActivationKey" in cloud_info:
                output += "ActivationKey : %s\n" % (cloud_info["ActivationKey"])
        if not json:
            self.rdmc.ui.printer(output, verbose_override=True)
        else:
            self.rdmc.ui.print_out_json(cloud_info)

    def run(self, line, help_disp=False):
        """Wrapper function for cloudconnect main function

        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.cmdbase.login_select_validation(self, options)
        if not self.rdmc.app.redfishinst:
            raise NoCurrentSessionEstablished("Please login to iLO and retry the command")

        ilo_ver = self.rdmc.app.getiloversion()
        if ilo_ver < 5.247:
            raise IncompatibleiLOVersionError(
                "ComputeOpsManagement Feature is only available with iLO 5 version 2.47 or higher.\n"
            )

        # validation checks
        self.cloudconnectvalidation(options)
        if options.command:
            if options.command.lower() == "connect":
                if options.proxy:
                    self.proxy_config(options.proxy)
                if options.activationkey:
                    self.connect_cloud(activationkey=options.activationkey)
                elif not options.activationkey:
                    self.connect_cloud()
                else:
                    raise InvalidCommandLineError(
                        "Activation Key %s is not alphanumeric or not of length 32." % str(options.activationkey)
                    )
            elif options.command.lower() == "disconnect":
                self.disconnect_cloud()
            elif options.command.lower() == "status":
                if options.json:
                    self.cloud_status(json=True)
                else:
                    self.cloud_status()
            else:
                raise InvalidCommandLineError("%s is not a valid option for this " "command." % str(options.command))
        else:
            raise InvalidCommandLineError(
                "Please provide either connect, disconnect or status as additional subcommand."
                " For help or usage related information, use -h or --help"
            )
        # logout routine
        self.cmdbase.logout_routine(self, options)
        # Return code
        return ReturnCodes.SUCCESS

    def cloudconnectvalidation(self, options):
        """new command method validation function"""
        # Check if Cloud Connect feature is enabled in iLO.
        path = self.rdmc.app.typepath.defs.managerpath
        resp = self.rdmc.app.get_handler(path, service=True, silent=True)
        if resp.status == 200:
            oem_actions = resp.dict["Oem"]["Hpe"]["Actions"]
            # print(oem_actions)
            if "#HpeiLO.EnableCloudConnect" not in oem_actions or "#HpeiLO.DisableCloudConnect" not in oem_actions:
                raise CloudConnectFailedError("ComputeOpsManagement is disabled in this iLO.\n")

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        subcommand_parser = customparser.add_subparsers(dest="command")
        connect_help = "To connect to ComputeOpsManagement\n"
        # connect sub-parser
        connect_parser = subcommand_parser.add_parser(
            "connect",
            help=connect_help,
            description=connect_help + "\n\tExample:\n\tcomputeopsmanagement connect or"
            "\n\tcomputeopsmanagement connect --proxy http://proxy.abc.com:8080 or "
            "\n\tcomputeopsmanagement connect --proxy None or "
            "\n\tcomputeopsmanagement connect --activationkey 123456789EFGA or "
            "\n\tcomputeopsmanagement connect --activationkey 123456789EFGA --proxy http://proxy.abc.com:8080 or "
            "\n\tcomputeopsmanagement connect --activationkey 123456789EFGA --proxy None",
            formatter_class=RawDescriptionHelpFormatter,
        )
        connect_parser.add_argument(
            "--activationkey",
            dest="activationkey",
            help="activation key is optional for connecting",
            required=False,
            type=str,
            default=None,
        )
        connect_parser.add_argument(
            "--proxy",
            dest="proxy",
            help="to set or clear proxy while connecting",
            type=str,
            default=None,
        )
        self.cmdbase.add_login_arguments_group(connect_parser)
        status_help = "To check the ComputeOpsManagement connection status\n"
        status_parser = subcommand_parser.add_parser(
            "status",
            help=status_help,
            description=status_help + "\n\tExample:\n\tcomputeopsmanagement status or "
            "\n\tcomputeopsmanagement status -j",
            formatter_class=RawDescriptionHelpFormatter,
        )
        status_parser.add_argument(
            "-j",
            "--json",
            dest="json",
            help="to print in json format",
            action="store_true",
            default=False,
        )
        self.cmdbase.add_login_arguments_group(status_parser)
        disconnect_help = "To disconnect from ComputeOpsManagement\n"
        disconnect_parser = subcommand_parser.add_parser(
            "disconnect",
            help=disconnect_help,
            description=disconnect_help + "\n\tExample:\n\tcomputeopsmanagement disconnect",
            formatter_class=RawDescriptionHelpFormatter,
        )
        self.cmdbase.add_login_arguments_group(disconnect_parser)



###
# Copyright 2021-2023 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
""" TwoFactorAuthentication Command for rdmc """

from argparse import RawDescriptionHelpFormatter

from redfish.ris import IdTokenError
from redfish.ris.rmc_helper import IloResponseError

try:
    from rdmc_helper import (
        IloLicenseError,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoCurrentSessionEstablished,
        ReturnCodes,
        TfaEnablePreRequisiteError,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        IloLicenseError,
        IncompatibleiLOVersionError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoCurrentSessionEstablished,
        ReturnCodes,
        TfaEnablePreRequisiteError,
    )

from redfish.ris.ris import SessionExpired


class SetTwoFactorAuthenticationCommand:
    """Main new command template class"""

    def __init__(self):
        self.ident = {
            "name": "settwofactorauthentication",
            "usage": "settwofactorauthentication\n\n",
            "description": "Run to enable ,disable ,get status or  setup smtp settings for Two factor "
            "authentication\n\t"
            "Example:\n\tsettwofactorauthentication enable or \n\t"
            "settwofactorauthentication disable \n\t"
            "settwofactorauthentication status or \n\t"
            "settwofactorauthentication status -j\n"
            "settwofactorauthentication smtp --smtpfortfaenabled true --alertmailsenderdomain 'test@hpe.com' "
            "--alertmailsmtpserver 'smtp3.hpe.com' --alertmailsmtpport 587\n",
            "summary": "Enables the server to use Two factor authentication, monitored",
            "aliases": [],
            "auxcommands": [],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def twofactorauth(self, options):
        """two authentication function

        :param enable or disable
        """

        # pre requisite checks

        path = self.rdmc.app.typepath.defs.managerpath + "NetworkProtocol"
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)

        status = resp.dict["Oem"]["Hpe"]["SMTPForTFAEnabled"]
        if not status:
            self.rdmc.ui.error("SMTP for TFA is not enabled\n")
            raise TfaEnablePreRequisiteError(
                "\nTo be able to enable or disable TFA\nKindly enable and set the SMTP details prior to enabling TFA "
                "for the server . You can use our command as given below for enabling  smtp for tfa \n\n"
                "SetTwoFactorAuthentication smtp --smtpfortfaenabled true --alertmailsenderdomain 'testuser@test.com' "
                "--alertmailsmtpserver 'smtp.server.com' \n"
            )

        results = self.rdmc.app.select(selector="AccountService.", path_refresh=True)[0].dict
        path = results[self.rdmc.app.typepath.defs.hrefstring]

        # check whether default schema is selected authentication tyep
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)
        if not resp.dict["LDAP"]["ServiceEnabled"]:
            self.rdmc.ui.error("Ldap service is not enabled\n")
            raise TfaEnablePreRequisiteError(
                "\nTo be able to enable or disable TFA\nKindly enable LDAP Directory Authentication to use Directory "
                "Default Schema. Same can be set using below command\n\ndirectory ldap enable "
                "--authentication=DefaultSchema\n"
            )

        body = dict()
        if options.command.lower() == "enable":
            serviceEnabled = "Enabled"
        else:
            serviceEnabled = "Disabled"
        body.update({"Oem": {"Hpe": {"TwoFactorAuth": serviceEnabled}}})

        try:
            self.rdmc.app.patch_handler(path, body)
            return ReturnCodes.SUCCESS
        except IloLicenseError:
            self.rdmc.ui.error("Error Occured while enable or disable operation for TFA")
            return ReturnCodes.ILO_LICENSE_ERROR
        except IdTokenError:
            self.rdmc.ui.error("Insufficient Privilege to enable or disable TFA ")
            return ReturnCodes.RIS_MISSING_ID_TOKEN
        except IloResponseError:
            self.rdmc.ui.error("Provided values are invalid , iLO threw error")
            return ReturnCodes.RIS_ILO_RESPONSE_ERROR

    def twofactorauth_smtp(self, options):
        """two authentication smtp setup function

        :param smtpfortfaenabled : true\false
        :type smtpfortfaenabled: str.
        """
        path = self.rdmc.app.typepath.defs.managerpath + "NetworkProtocol"
        resp_get = self.rdmc.app.get_handler(path, service=False, silent=True)
        body = dict()
        range1 = range(1, 65536, 1)
        if options.SMTPPort is not None:
            if options.SMTPPort in range1:
                SMTPPort = options.SMTPPort
            else:
                raise InvalidCommandLineError("Enter correct server port value in the range 1 to 65535")
        else:
            SMTPPort = 25

        if options.SMTPForTFA[0].lower() == "true":
            SMTPForTFAEnabled = True
        else:
            SMTPForTFAEnabled = False

        if options.SenderDomain is None:
            if resp_get.dict["Oem"]["Hpe"]["AlertMailSenderDomain"] is not None:
                SenderDomain = resp_get.dict["Oem"]["Hpe"]["AlertMailSenderDomain"]
            else:
                SenderDomain = ""
        else:
            SenderDomain = options.SenderDomain[0]

        if options.SMTPServer is None:
            if resp_get.dict["Oem"]["Hpe"]["AlertMailSMTPServer"] is not None:
                SMTPServer = resp_get.dict["Oem"]["Hpe"]["AlertMailSMTPServer"]
            else:
                SMTPServer = ""
        else:
            SMTPServer = options.SMTPServer[0]

        body.update(
            {
                "Oem": {
                    "Hpe": {
                        "SMTPForTFAEnabled": SMTPForTFAEnabled,
                        "AlertMailSenderDomain": SenderDomain,
                        "AlertMailSMTPServer": SMTPServer,
                        "AlertMailSMTPPort": SMTPPort,
                    }
                }
            }
        )
        try:
            self.rdmc.app.patch_handler(path, body)
            return ReturnCodes.SUCCESS

        except IloResponseError:
            self.rdmc.ui.error("Provided values are invalid , iLO threw error")
            return ReturnCodes.RIS_ILO_RESPONSE_ERROR
        except IloLicenseError:
            self.rdmc.ui.error("Error Occured while setting smtp settings of TFA ")
            return ReturnCodes.ILO_LICENSE_ERROR
        except IdTokenError:
            self.rdmc.ui.error("Insufficient Privilege to setting smtp settings of TFA ")
            return ReturnCodes.RIS_MISSING_ID_TOKEN

    def twofactorauth_status(self, json=False):
        """two authentication status function

        :param json: json
        :type json: bool
        """
        results = self.rdmc.app.select(selector="AccountService.", path_refresh=True)[0].dict
        path = results[self.rdmc.app.typepath.defs.hrefstring]
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)
        if resp.status != 200:
            self.rdmc.ui.error("session seems to be expired\n")
            raise SessionExpired("Invalid session. Please logout and log back in or include credentials.")
        tfa_info = resp.dict["Oem"]["Hpe"]
        output = "------------------------------------------------\n"
        output += "TFA Status : %s\n" % (tfa_info["TwoFactorAuth"])
        if not json:
            self.rdmc.ui.printer(output, verbose_override=True)
        else:
            self.rdmc.ui.printer("\n--- TFA status ---\n")
            self.rdmc.ui.print_out_json(tfa_info)

        path = self.rdmc.app.typepath.defs.managerpath + "NetworkProtocol"
        resp = self.rdmc.app.get_handler(path, service=False, silent=True)
        if resp.status != 200:
            raise SessionExpired("Invalid session. Please logout and log back in or include credentials.")
        smtp_info = resp.dict["Oem"]["Hpe"]
        output = "\n------------------------------------------------\n"
        output += "SMTP for TFA Status : %s\n" % (smtp_info["SMTPForTFAEnabled"])
        if not json:
            self.rdmc.ui.printer(output, verbose_override=True)
        else:
            self.rdmc.ui.printer("\n--- SMTP for TFA status ---\n")
            self.rdmc.ui.print_out_json(smtp_info)

        return ReturnCodes.SUCCESS

    def run(self, line, help_disp=False):
        """Wrapper function for two factor authentication main function

        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            line.append("-h")
            try:
                (_, _) = self.rdmc.rdmc_parse_arglist(self, line)
            except:
                return ReturnCodes.SUCCESS
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.cmdbase.login_select_validation(self, options)
        if not self.rdmc.app.redfishinst:
            raise NoCurrentSessionEstablished("Please login to iLO and retry the command")

        ilo_ver = self.rdmc.app.getiloversion()
        if ilo_ver < 5.290 or (ilo_ver < 6.150 and ilo_ver > 5.999):
            raise IncompatibleiLOVersionError(
                "TFA Feature is only available with iLO 5 version 2.90 or higher and iLO 6 version 1.50 or higher.\n"
            )

        # validation checks
        self.twofactorauthenticationvalidation(options)
        if options.command:
            if options.command.lower() == "enable":
                returncode = self.twofactorauth(options)
            elif options.command.lower() == "disable":
                returncode = self.twofactorauth(options)
            elif options.command.lower() == "smtp":
                returncode = self.twofactorauth_smtp(options)
            elif options.command.lower() == "status":
                if options.json:
                    returncode = self.twofactorauth_status(json=True)
                else:
                    returncode = self.twofactorauth_status()
            else:
                raise InvalidCommandLineError("%s is not a valid option for this " "command." % str(options.command))
        else:
            raise InvalidCommandLineError(
                "Please provide either enable, disable , status or smtp as additional subcommand. "
                "For help or usage related information, use -h or --help"
            )
        # logout routine
        self.cmdbase.logout_routine(self, options)
        # Return code
        return returncode

    def twofactorauthenticationvalidation(self, options):
        """new command method validation function"""
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)
        subcommand_parser = customparser.add_subparsers(dest="command")
        enable_help = "To enable two factor authentication\n"
        # connect sub-parser
        enable_parser = subcommand_parser.add_parser(
            "enable",
            help=enable_help,
            description=enable_help + "\n\tExample:\n\tsettwofactorauthentication enable ",
            formatter_class=RawDescriptionHelpFormatter,
        )

        self.cmdbase.add_login_arguments_group(enable_parser)

        status_help = "To check the settwofactorauthentication status\n"
        status_parser = subcommand_parser.add_parser(
            "status",
            help=status_help,
            description=status_help + "\n\tExample:\n\tsettwofactorauthentication status or "
            "\n\tsettwofactorauthentication status -j",
            formatter_class=RawDescriptionHelpFormatter,
        )
        status_parser.add_argument(
            "-j",
            "--json",
            dest="json",
            help="to print in json format",
            action="store_true",
            default=False,
        )
        self.cmdbase.add_login_arguments_group(status_parser)

        disable_help = "To disable the two factor authentication\n"
        disable_parser = subcommand_parser.add_parser(
            "disable",
            help=disable_help,
            description=disable_help + "\n\tExample:\n\tsettwofactorauthentication disable",
            formatter_class=RawDescriptionHelpFormatter,
        )
        self.cmdbase.add_login_arguments_group(disable_parser)

        emailconfig_help = "To set the smtp setting for settwofactorauthentication\n"
        emailconfig_parser = subcommand_parser.add_parser(
            "smtp",
            help=emailconfig_help,
            description=emailconfig_help
            + "\n\tExample:\n\tsettwofactorauthentication smtp --smtpfortfaenabled true --alertmailsenderdomain "
            "'test@hpe.com' --alertmailsmtpserver 'smtp3.hpe.com' --alertmailsmtpport 587 ",
            formatter_class=RawDescriptionHelpFormatter,
        )
        emailconfig_parser.add_argument(
            "--smtpfortfaenabled",
            dest="SMTPForTFA",
            help="smtp service enable , supply either true or false as input. "
            "If true then remaining inputs cant be null or empty",
            action="append",
            default=None,
        )
        emailconfig_parser.add_argument(
            "--alertmailsenderdomain",
            dest="SenderDomain",
            action="append",
            help="smtp user name , supply a valid aduser for this field",
            default=None,
        )
        emailconfig_parser.add_argument(
            "--alertmailsmtpserver",
            dest="SMTPServer",
            action="append",
            help="smtp server , supply a valid smtp server for this parameter",
            default=None,
        )

        emailconfig_parser.add_argument(
            "--alertmailsmtpport",
            dest="SMTPPort",
            type=int,
            help="smtp port , supply a valid smtp port for this parameter, default is 25",
            default=25,
        )

        self.cmdbase.add_login_arguments_group(emailconfig_parser)



from __future__ import absolute_import

import sys
import time
from contextlib import contextmanager

import paramiko


class IloCli(object):
    """Wrapper for the iLO CLI (SMASH CLP)."""

    def __init__(self, hostname, username=None, password=None, port=22, timeout=10, window_size=None):
        """Initialize an iLO CLI instance.

        :param hostname: iLO hostname
        :param username: iLO username
        :param password: iLO password
        :param port: SSH port
        :param window_size: window size for the SSH session (max is used if not specified)
        :param timeout: connection timeout, in seconds
        """
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.timeout = timeout
        self.window_size = window_size if window_size else paramiko.common.MAX_WINDOW_SIZE
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.channel = None

    def open(self):
        """Start an interactive iLO CLI session."""
        self.ssh.connect(self.hostname, username=self.username, password=self.password,
                         port=self.port, timeout=self.timeout, look_for_keys=False)
        transport = self.ssh.get_transport()
        self.channel = transport.open_session(window_size=self.window_size)
        self.channel.get_pty()
        self.channel.invoke_shell()

    def close(self):
        """End the iLO CLI session."""
        if self.channel:
            self.channel.close()
        self.ssh.close()

    def __enter__(self):
        for _ in range(5):
            try:
                self.open()
                return self
            except paramiko.SSHException:
                sys.stdout.write("SSHException: retrying connection")
                time.sleep(30)

        sys.stdout.write("SSH: Max retries attempted. No vsp will be captured.")
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def read(self):
        """Read all available output without blocking I/O."""
        buf = []
        while self.channel.recv_ready():
            data = self.channel.recv(1024)
            buf.append(data)
        return "".join(buf)

    def send(self, data):
        self.channel.send(data)

    def sendline(self, data):
        self.send(data + '\n')

    @contextmanager
    def vsp_session(self):
        """Start a Virtual Serial Port session, returning to the CLI afterward."""
        self.sendline("vsp")
        yield self
        self.send("\x1B(")  # press 'ESC ('
